"""Tests related to repository versions."""
import unittest
import pytest
from random import choice, randint, sample
from time import sleep
from urllib.parse import urlsplit
from tempfile import NamedTemporaryFile
from hashlib import sha256

from pulp_smash import api, config, utils
from pulp_smash.exceptions import TaskReportError
from pulp_smash.pulp3.bindings import delete_orphans, monitor_task
from pulp_smash.pulp3.constants import ARTIFACTS_PATH
from pulp_smash.pulp3.utils import (
    download_content_unit,
    delete_version,
    gen_repo,
    gen_distribution,
    get_added_content,
    get_added_content_summary,
    get_artifact_paths,
    get_content,
    get_content_summary,
    get_removed_content,
    get_removed_content_summary,
    get_versions,
    modify_repo,
    sync,
)
from requests.exceptions import HTTPError

from pulpcore.client.pulpcore import ApiClient as CoreApiClient
from pulpcore.client.pulpcore import RepositoryVersionsApi
from pulpcore.client.pulp_file import (
    ContentFilesApi,
    DistributionsFileApi,
    PublicationsFileApi,
    RemotesFileApi,
    RepositoriesFileApi,
    RepositoriesFileVersionsApi,
    RepositorySyncURL,
)
from pulpcore.client.pulp_file.exceptions import ApiException
from pulpcore.tests.functional.api.using_plugin.constants import (
    FILE2_FIXTURE_MANIFEST_URL,
    FILE_CONTENT_NAME,
    FILE_CONTENT_PATH,
    FILE_DISTRIBUTION_PATH,
    FILE_FIXTURE_COUNT,
    FILE_FIXTURE_MANIFEST_URL,
    FILE_FIXTURE_SUMMARY,
    FILE_LARGE_FIXTURE_MANIFEST_URL,
    FILE_MANY_FIXTURE_MANIFEST_URL,
    FILE_REMOTE_PATH,
    FILE_REPO_PATH,
    FILE_URL,
    FILE2_URL,
)
from pulpcore.tests.functional.api.using_plugin.utils import (
    create_distribution,
    create_file_publication,
    gen_file_client,
    gen_file_remote,
    populate_pulp,
)
from pulpcore.tests.functional.api.using_plugin.utils import set_up_module as setUpModule  # noqa
from pulpcore.tests.functional.api.using_plugin.utils import skip_if


def remove_created_key(dic):
    """Given a dict remove the key `created`."""
    return {k: v for k, v in dic.items() if k != "created"}


class AddRemoveContentTestCase(unittest.TestCase):
    """Add and remove content to a repository. Verify side-effects.

    A new repository version is automatically created each time content is
    added to or removed from a repository. Furthermore, it's possible to
    inspect any repository version and discover which content is present, which
    content was removed, and which content was added. This test case explores
    these features.

    This test targets the following issues:

    * `Pulp #3059 <https://pulp.plan.io/issues/3059>`_
    * `Pulp #3234 <https://pulp.plan.io/issues/3234>`_
    * `Pulp Smash #878 <https://github.com/pulp/pulp-smash/issues/878>`_
    """

    @classmethod
    def setUpClass(cls):
        """Create class-wide variables."""
        cls.cfg = config.get_config()
        cls.client = api.Client(cls.cfg, api.page_handler)
        cls.remote = {}
        cls.repo = {}
        cls.content = {}

    @classmethod
    def tearDownClass(cls):
        """Destroy resources created by test methods."""
        if cls.remote:
            cls.client.delete(cls.remote["pulp_href"])
        if cls.repo:
            cls.client.delete(cls.repo["pulp_href"])

    def test_01_create_repository(self):
        """Create a repository.

        Assert that:

        * The ``versions_href`` API call is correct.
        * The ``latest_version_href`` API call is correct.
        """
        self.repo.update(self.client.post(FILE_REPO_PATH, gen_repo()))

        repo_versions = get_versions(self.repo)
        self.assertEqual(len(repo_versions), 1, repo_versions)

        self.assertEqual(self.repo["latest_version_href"], f"{self.repo['pulp_href']}versions/0/")

    @skip_if(bool, "repo", False)
    def test_02_sync_content(self):
        """Sync content into the repository.

        Assert that:

        * The ``versions_href`` API call is correct.
        * The ``latest_version_href`` API call is correct.
        * The ``content_hrefs`` attribute is correct.
        * The ``content_added_hrefs`` attribute is correct.
        * The ``content_removed_hrefs`` attribute is correct.
        * The ``content_summary`` attribute is correct.
        * The ``content_added_summary`` attribute is correct.
        * The ``content_removed_summary`` attribute is correct.
        """
        body = gen_file_remote()
        body.update({"headers": [{"Connection": "keep-alive"}]})
        self.remote.update(self.client.post(FILE_REMOTE_PATH, body))
        sync(self.cfg, self.remote, self.repo)
        repo = self.client.get(self.repo["pulp_href"])

        repo_versions = get_versions(repo)
        self.assertEqual(len(repo_versions), 2, repo_versions)

        self.assertIsNotNone(repo["latest_version_href"])

        content_hrefs = get_content(repo)[FILE_CONTENT_NAME]
        self.assertEqual(len(content_hrefs), FILE_FIXTURE_COUNT, content_hrefs)

        content = get_content(repo)[FILE_CONTENT_NAME]
        self.assertEqual(len(content), FILE_FIXTURE_COUNT)

        content_added = get_added_content(repo)[FILE_CONTENT_NAME]
        self.assertEqual(len(content_added), FILE_FIXTURE_COUNT)

        content_removed = get_removed_content(repo)[FILE_CONTENT_NAME]
        self.assertEqual(len(content_removed), 0)

        content_summary = get_content_summary(repo)
        self.assertDictEqual(content_summary, FILE_FIXTURE_SUMMARY)

        content_added_summary = get_added_content_summary(repo)
        self.assertDictEqual(content_added_summary, FILE_FIXTURE_SUMMARY)

        content_removed_summary = get_removed_content_summary(repo)
        self.assertDictEqual(content_removed_summary, {})

    @skip_if(bool, "repo", False)
    def test_03_remove_content(self):
        """Remove content from the repository.

        Make roughly the same assertions as :meth:`test_02_sync_content`.
        """
        repo = self.client.get(self.repo["pulp_href"])
        self.content.update(choice(get_content(repo)[FILE_CONTENT_NAME]))

        modify_repo(self.cfg, self.repo, remove_units=[self.content])
        repo = self.client.get(self.repo["pulp_href"])

        repo_versions = get_versions(repo)
        self.assertEqual(len(repo_versions), 3, repo_versions)

        self.assertIsNotNone(repo["latest_version_href"])

        content_hrefs = get_content(repo)[FILE_CONTENT_NAME]
        self.assertEqual(len(content_hrefs), FILE_FIXTURE_COUNT - 1, content_hrefs)

        content = get_content(repo)[FILE_CONTENT_NAME]
        self.assertEqual(len(content), FILE_FIXTURE_COUNT - 1)

        added_content = get_added_content(repo)[FILE_CONTENT_NAME]
        self.assertListEqual(added_content, [], added_content)

        removed_content = get_removed_content(repo)[FILE_CONTENT_NAME]
        self.assertListEqual(removed_content, [self.content], removed_content)

        content_summary = get_content_summary(repo)
        self.assertDictEqual(content_summary, {FILE_CONTENT_NAME: FILE_FIXTURE_COUNT - 1})

        content_added_summary = get_added_content_summary(repo)
        self.assertDictEqual(content_added_summary, {})

        content_removed_summary = get_removed_content_summary(repo)
        self.assertDictEqual(content_removed_summary, {FILE_CONTENT_NAME: 1})

    @skip_if(bool, "repo", False)
    def test_04_add_content(self):
        """Add content to the repository.

        Make roughly the same assertions as :meth:`test_02_sync_content`.
        """
        repo = self.client.get(self.repo["pulp_href"])
        modify_repo(self.cfg, self.repo, add_units=[self.content])
        repo = self.client.get(self.repo["pulp_href"])

        repo_versions = get_versions(repo)
        self.assertEqual(len(repo_versions), 4, repo_versions)

        self.assertIsNotNone(repo["latest_version_href"])

        content_hrefs = get_content(repo)[FILE_CONTENT_NAME]
        self.assertEqual(len(content_hrefs), FILE_FIXTURE_COUNT, content_hrefs)

        content = get_content(repo)[FILE_CONTENT_NAME]
        self.assertEqual(len(content), FILE_FIXTURE_COUNT)

        added_content = get_added_content(repo)[FILE_CONTENT_NAME]
        self.assertListEqual(added_content, [self.content], added_content)

        removed_content = get_removed_content(repo)[FILE_CONTENT_NAME]
        self.assertListEqual(removed_content, [], removed_content)

        content_summary = get_content_summary(repo)
        self.assertDictEqual(content_summary, FILE_FIXTURE_SUMMARY)

        content_added_summary = get_added_content_summary(repo)
        self.assertDictEqual(content_added_summary, {FILE_CONTENT_NAME: 1})

        content_removed_summary = get_removed_content_summary(repo)
        self.assertDictEqual(content_removed_summary, {})

    def get_content_summary(self, repo):
        """Get the ``content_summary`` for the given repository."""
        repo_versions = get_versions(repo)
        content_summaries = [
            repo_version["content_summary"]
            for repo_version in repo_versions
            if repo_version["pulp_href"] == repo["latest_version_href"]
        ]
        self.assertEqual(len(content_summaries), 1, content_summaries)
        return content_summaries[0]


class SyncChangeRepoVersionTestCase(unittest.TestCase):
    """Verify whether sync of repository updates repository version."""

    def test_all(self):
        """Verify whether the sync of a repository updates its version.

        This test explores the design choice stated in the `Pulp #3308`_ that a
        new repository version is created even if the sync does not add or
        remove any content units. Even without any changes to the remote if a
        new sync occurs, a new repository version is created.

        .. _Pulp #3308: https://pulp.plan.io/issues/3308

        Do the following:

        1. Create a repository, and a remote.
        2. Sync the repository.
        3. Remove all content - one by one.
        3. Verify that the repository version is equal to the number of operations.
        """
        cfg = config.get_config()
        client = api.Client(cfg, api.json_handler)

        repo = client.post(FILE_REPO_PATH, gen_repo())
        self.addCleanup(client.delete, repo["pulp_href"])

        body = gen_file_remote()
        remote = client.post(FILE_REMOTE_PATH, body)
        self.addCleanup(client.delete, remote["pulp_href"])

        sync(cfg, remote, repo)
        repo = client.get(repo["pulp_href"])
        for file_content in get_content(repo)[FILE_CONTENT_NAME]:
            modify_repo(cfg, repo, remove_units=[file_content])
        repo = client.get(repo["pulp_href"])
        path = urlsplit(repo["latest_version_href"]).path
        latest_repo_version = int(path.split("/")[-2])
        self.assertEqual(latest_repo_version, 4)


class AddRemoveRepoVersionTestCase(unittest.TestCase):
    """Create and delete repository versions.

    This test targets the following issues:

    * `Pulp #3219 <https://pulp.plan.io/issues/3219>`_
    * `Pulp Smash #871 <https://github.com/pulp/pulp-smash/issues/871>`_
    """

    # `cls.content[i]` is a dict.
    # pylint:disable=unsubscriptable-object

    @classmethod
    def setUpClass(cls):
        """Add content to Pulp."""
        cls.cfg = config.get_config()
        cls.client = api.Client(cls.cfg, api.json_handler)
        delete_orphans()
        populate_pulp(cls.cfg, url=FILE_LARGE_FIXTURE_MANIFEST_URL)
        # We need at least three content units. Choosing a relatively low
        # number is useful, to limit how many repo versions are created, and
        # thus how long the test takes.
        cls.content = sample(cls.client.get(FILE_CONTENT_PATH)["results"], 4)

    def setUp(self):
        """Create a repository and give it nine new versions."""
        self.repo = self.client.post(FILE_REPO_PATH, gen_repo())
        self.addCleanup(self.client.delete, self.repo["pulp_href"])

        # Don't upload the last content unit. The test case might upload it to
        # create a new repo version within the test.
        for content in self.content[:-1]:
            self.client.post(
                self.repo["pulp_href"] + "modify/", {"add_content_units": [content["pulp_href"]]}
            )
        self.repo = self.client.get(self.repo["pulp_href"])
        self.repo_version_hrefs = tuple(version["pulp_href"] for version in get_versions(self.repo))

    def test_delete_first_version(self):
        """Delete the first repository version (version 0)."""
        delete_version(self.repo, self.repo_version_hrefs[0])

    def test_delete_last_version(self):
        """Delete the last repository version.

        Create a new repository version from the second-to-last repository
        version. Verify that the content unit from the old last repository
        version is not in the new last repository version.
        """
        # Delete the last repo version.
        delete_version(self.repo, self.repo_version_hrefs[-1])
        with self.assertRaises(HTTPError):
            get_content(self.repo, self.repo_version_hrefs[-1])

        # Make new repo version from new last repo version.
        self.client.post(
            self.repo["pulp_href"] + "modify/",
            {"add_content_units": [self.content[-1]["pulp_href"]]},
        )
        self.repo = self.client.get(self.repo["pulp_href"])
        artifact_paths = get_artifact_paths(self.repo)

        self.assertNotIn(self.content[-2]["artifact"], artifact_paths)
        self.assertIn(self.content[-1]["artifact"], artifact_paths)

    def test_delete_middle_version(self):
        """Delete a middle version."""
        index = randint(1, len(self.repo_version_hrefs) - 3)
        delete_version(self.repo, self.repo_version_hrefs[index])

        with self.assertRaises(HTTPError):
            get_content(self.repo, self.repo_version_hrefs[index])

        # Check added count is updated properly
        added = get_added_content_summary(self.repo, self.repo_version_hrefs[index + 1])
        self.assertEqual(added["file.file"], 2)

        for repo_version_href in self.repo_version_hrefs[index + 1 :]:
            artifact_paths = get_artifact_paths(self.repo, repo_version_href)
            self.assertIn(self.content[index]["artifact"], artifact_paths)

    def test_delete_all_versions(self):
        """Attempt to delete all versions."""
        for repo_version_href in self.repo_version_hrefs[:-1]:
            delete_version(self.repo, repo_version_href)

        with self.assertRaises(TaskReportError) as ctx:
            delete_version(self.repo, self.repo_version_hrefs[-1])

        self.assertIn(
            "Cannot delete repository version.", ctx.exception.task["error"]["description"]
        )

    def test_delete_publication(self):
        """Delete a publication.

        Delete a repository version, and verify the associated publication is
        also deleted.
        """
        publication = create_file_publication(self.cfg, self.repo)
        delete_version(self.repo)

        with self.assertRaises(HTTPError):
            self.client.get(publication["pulp_href"])


@pytest.mark.parallel
def test_squash_repo_version(
    file_repo_api_client, file_repo_version_api_client, content_file_api_client, file_repo
):
    """Test that the deletion of a repository version properly squashes the content.

    - Setup versions like:
        Version 0: <empty>
            add: ABCDE
        Version 1: ABCDE
            delete: BCDE; add: FGHI
        Version 2: AFGHI -- to be deleted
            delete: GI; add: CD
        Version 3: ACDFH -- to be squashed into
            delete: DH; add: EI
        Version 4: ACEFI
    - Delete version 2.
    - Check the content of all remaining versions.
    """
    content_units = {}
    for name in ["A", "B", "C", "D", "E", "F", "G", "H", "I"]:
        try:
            content_units[name] = content_file_api_client.list(
                relative_path=name, sha256=sha256(name.encode()).hexdigest()
            ).results[0]
        except IndexError:
            with NamedTemporaryFile() as tf:
                tf.write(name.encode())
                tf.flush()
                response = content_file_api_client.create(relative_path=name, file=tf.name)
                result = monitor_task(response.task)
                content_units[name] = content_file_api_client.read(result.created_resources[0])
    response1 = file_repo_api_client.modify(
        file_repo.pulp_href,
        {
            "add_content_units": [
                content.pulp_href
                for key, content in content_units.items()
                if key in ["A", "B", "C", "D", "E"]
            ]
        },
    )

    response2 = file_repo_api_client.modify(
        file_repo.pulp_href,
        {
            "remove_content_units": [
                content.pulp_href
                for key, content in content_units.items()
                if key in ["B", "C", "D", "E"]
            ],
            "add_content_units": [
                content.pulp_href
                for key, content in content_units.items()
                if key in ["F", "G", "H", "I"]
            ],
        },
    )

    response3 = file_repo_api_client.modify(
        file_repo.pulp_href,
        {
            "remove_content_units": [
                content.pulp_href for key, content in content_units.items() if key in ["G", "I"]
            ],
            "add_content_units": [
                content.pulp_href for key, content in content_units.items() if key in ["C", "D"]
            ],
        },
    )

    response4 = file_repo_api_client.modify(
        file_repo.pulp_href,
        {
            "remove_content_units": [
                content.pulp_href for key, content in content_units.items() if key in ["D", "H"]
            ],
            "add_content_units": [
                content.pulp_href for key, content in content_units.items() if key in ["E", "I"]
            ],
        },
    )
    version1 = file_repo_version_api_client.read(monitor_task(response1.task).created_resources[0])
    version2 = file_repo_version_api_client.read(monitor_task(response2.task).created_resources[0])
    version3 = file_repo_version_api_client.read(monitor_task(response3.task).created_resources[0])
    version4 = file_repo_version_api_client.read(monitor_task(response4.task).created_resources[0])

    # Check version state before deletion
    assert version1.content_summary.added["file.file"]["count"] == 5
    assert "file.file" not in version1.content_summary.removed
    assert version2.content_summary.added["file.file"]["count"] == 4
    assert version2.content_summary.removed["file.file"]["count"] == 4
    assert version3.content_summary.added["file.file"]["count"] == 2
    assert version3.content_summary.removed["file.file"]["count"] == 2
    assert version4.content_summary.added["file.file"]["count"] == 2
    assert version4.content_summary.removed["file.file"]["count"] == 2

    content1 = content_file_api_client.list(repository_version=version1.pulp_href)
    content2 = content_file_api_client.list(repository_version=version2.pulp_href)
    content3 = content_file_api_client.list(repository_version=version3.pulp_href)
    content4 = content_file_api_client.list(repository_version=version4.pulp_href)
    assert set((content.relative_path for content in content1.results)) == {"A", "B", "C", "D", "E"}
    assert set((content.relative_path for content in content2.results)) == {"A", "F", "G", "H", "I"}
    assert set((content.relative_path for content in content3.results)) == {"A", "C", "D", "F", "H"}
    assert set((content.relative_path for content in content4.results)) == {"A", "C", "E", "F", "I"}

    monitor_task(file_repo_version_api_client.delete(version2.pulp_href).task)

    # Check version state after deletion (Version 2 is gone...)
    version1 = file_repo_version_api_client.read(version1.pulp_href)
    version3 = file_repo_version_api_client.read(version3.pulp_href)
    version4 = file_repo_version_api_client.read(version4.pulp_href)

    assert version1.content_summary.added["file.file"]["count"] == 5
    assert "file.file" not in version1.content_summary.removed
    assert version3.content_summary.added["file.file"]["count"] == 2
    assert version3.content_summary.removed["file.file"]["count"] == 2
    assert version4.content_summary.added["file.file"]["count"] == 2
    assert version4.content_summary.removed["file.file"]["count"] == 2

    content1 = content_file_api_client.list(repository_version=version1.pulp_href)
    content3 = content_file_api_client.list(repository_version=version3.pulp_href)
    content4 = content_file_api_client.list(repository_version=version4.pulp_href)
    assert set((content.relative_path for content in content1.results)) == {"A", "B", "C", "D", "E"}
    assert set((content.relative_path for content in content3.results)) == {"A", "C", "D", "F", "H"}
    assert set((content.relative_path for content in content4.results)) == {"A", "C", "E", "F", "I"}


class ContentImmutableRepoVersionTestCase(unittest.TestCase):
    """Test whether the content present in a repo version is immutable.

    This test targets the following issue:

    * `Pulp Smash #953 <https://github.com/pulp/pulp-smash/issues/953>`_
    """

    def test_all(self):
        """Test whether the content present in a repo version is immutable.

        Do the following:

        1. Create a repository that has at least one repository version.
        2. Attempt to update the content of a repository version.
        3. Assert that an HTTP exception is raised.
        4. Assert that the repository version was not updated.
        """
        cfg = config.get_config()
        client = api.Client(cfg, api.json_handler)

        repo = client.post(FILE_REPO_PATH, gen_repo())
        self.addCleanup(client.delete, repo["pulp_href"])

        body = gen_file_remote()
        remote = client.post(FILE_REMOTE_PATH, body)
        self.addCleanup(client.delete, remote["pulp_href"])

        sync(cfg, remote, repo)

        latest_version_href = client.get(repo["pulp_href"])["latest_version_href"]
        with self.assertRaises(HTTPError):
            client.post(latest_version_href)
        repo = client.get(repo["pulp_href"])
        self.assertEqual(latest_version_href, repo["latest_version_href"])


class FilterRepoVersionTestCase(unittest.TestCase):
    """Test whether repository versions can be filtered.

    These tests target the following issues:

    * `Pulp #3238 <https://pulp.plan.io/issues/3238>`_
    * `Pulp #3536 <https://pulp.plan.io/issues/3536>`_
    * `Pulp #3557 <https://pulp.plan.io/issues/3557>`_
    * `Pulp #3558 <https://pulp.plan.io/issues/3558>`_
    * `Pulp Smash #880 <https://github.com/pulp/pulp-smash/issues/880>`_
    """

    @classmethod
    def setUpClass(cls):
        """Create class-wide variables.

        Add content to Pulp.
        """
        cls.cfg = config.get_config()
        cls.client = api.Client(cls.cfg, api.json_handler)

        populate_pulp(cls.cfg)
        cls.contents = cls.client.get(FILE_CONTENT_PATH)["results"]

    def setUp(self):
        """Create a repository and give it new versions."""
        self.repo = self.client.post(FILE_REPO_PATH, gen_repo())
        self.addCleanup(self.client.delete, self.repo["pulp_href"])

        for content in self.contents[:10]:  # slice is arbitrary upper bound
            modify_repo(self.cfg, self.repo, add_units=[content])
            sleep(1)
        self.repo = self.client.get(self.repo["pulp_href"])

    def test_filter_invalid_content(self):
        """Filter repository version by invalid content."""
        with self.assertRaises(HTTPError):
            get_versions(self.repo, {"content": utils.uuid4()})

    def test_filter_valid_content(self):
        """Filter repository versions by valid content."""
        content = choice(self.contents)
        repo_versions = get_versions(self.repo, {"content": content["pulp_href"]})
        for repo_version in repo_versions:
            self.assertIn(
                self.client.get(content["pulp_href"]),
                get_content(self.repo, repo_version["pulp_href"])[FILE_CONTENT_NAME],
            )

    def test_filter_invalid_date(self):
        """Filter repository version by invalid date."""
        criteria = utils.uuid4()
        for params in (
            {"pulp_created": criteria},
            {"pulp_created__gt": criteria, "pulp_created__lt": criteria},
            {"pulp_created__gte": criteria, "pulp_created__lte": criteria},
            {"pulp_created__range": ",".join((criteria, criteria))},
        ):
            with self.subTest(params=params):
                with self.assertRaises(HTTPError):
                    get_versions(self.repo, params)

    def test_filter_valid_date(self):
        """Filter repository version by a valid date."""
        dates = self.get_repo_versions_attr("pulp_created")
        for params, num_results in (
            ({"pulp_created": dates[0]}, 1),
            ({"pulp_created__gt": dates[0], "pulp_created__lt": dates[-1]}, len(dates) - 2),
            ({"pulp_created__gte": dates[0], "pulp_created__lte": dates[-1]}, len(dates)),
            ({"pulp_created__range": ",".join((dates[0], dates[1]))}, 2),
        ):
            with self.subTest(params=params):
                results = get_versions(self.repo, params)
                self.assertEqual(len(results), num_results, results)

    def test_filter_nonexistent_version(self):
        """Filter repository version by a nonexistent version number."""
        criteria = -1
        for params in (
            {"number": criteria},
            {"number__gt": criteria, "number__lt": criteria},
            {"number__gte": criteria, "number__lte": criteria},
            {"number__range": ",".join((str(criteria), str(criteria)))},
        ):
            with self.subTest(params=params):
                versions = get_versions(self.repo, params)
                self.assertEqual(len(versions), 0, versions)

    def test_filter_invalid_version(self):
        """Filter repository version by an invalid version number."""
        criteria = utils.uuid4()
        for params in (
            {"number": criteria},
            {"number__gt": criteria, "number__lt": criteria},
            {"number__gte": criteria, "number__lte": criteria},
            {"number__range": ",".join((criteria, criteria))},
        ):
            with self.subTest(params=params):
                with self.assertRaises(HTTPError):
                    get_versions(self.repo, params)

    def test_filter_valid_version(self):
        """Filter repository version by a valid version number."""
        numbers = self.get_repo_versions_attr("number")
        for params, num_results in (
            ({"number": numbers[0]}, 1),
            ({"number__gt": numbers[0], "number__lt": numbers[-1]}, len(numbers) - 2),
            ({"number__gte": numbers[0], "number__lte": numbers[-1]}, len(numbers)),
            ({"number__range": "{},{}".format(numbers[0], numbers[1])}, 2),
        ):
            with self.subTest(params=params):
                results = get_versions(self.repo, params)
                self.assertEqual(len(results), num_results, results)

    def test_deleted_version_filter(self):
        """Delete a repository version and filter by its number."""
        numbers = self.get_repo_versions_attr("number")
        delete_version(self.repo)
        versions = get_versions(self.repo, {"number": numbers[-1]})
        self.assertEqual(len(versions), 0, versions)

    def get_repo_versions_attr(self, attr):
        """Get an ``attr`` about each version of ``self.repo``.

        Return as sorted list.
        """
        attributes = [version[attr] for version in get_versions(self.repo)]
        attributes.sort()
        return attributes


class CreatedResourcesTaskTestCase(unittest.TestCase):
    """Verify whether task report shows that a repository version was created.

    This test targets the following issue:

    `Pulp Smash #876 <https://github.com/pulp/pulp-smash/issues/876>`_.
    """

    def test_all(self):
        """Verify whether task report shows repository version was created."""
        cfg = config.get_config()
        client = api.Client(cfg, api.json_handler)

        repo = client.post(FILE_REPO_PATH, gen_repo())
        self.addCleanup(client.delete, repo["pulp_href"])

        body = gen_file_remote()
        remote = client.post(FILE_REMOTE_PATH, body)
        self.addCleanup(client.delete, remote["pulp_href"])

        call_report = sync(cfg, remote, repo)
        for key in ("repositories", "versions"):
            self.assertIn(key, call_report["pulp_href"], call_report)


class CreateRepoBaseVersionTestCase(unittest.TestCase):
    """Test whether one can create a repository version from any version.

    This test targets the following issues:

    `Pulp #3360 <https://pulp.plan.io/issues/3360>`_
    `Pulp #4035 <https://pulp.plan.io/issues/4035>`_
    """

    @classmethod
    def setUpClass(cls):
        """Create class-wide variables."""
        cls.cfg = config.get_config()
        delete_orphans()
        populate_pulp(cls.cfg, url=FILE_LARGE_FIXTURE_MANIFEST_URL)
        cls.client = api.Client(cls.cfg, api.page_handler)
        cls.content = cls.client.get(FILE_CONTENT_PATH)

    def test_same_repository(self):
        """Test ``base_version`` for the same repository.

        Do the following:

        1. Create a repository.
        2. Sync the repository (this creates repository version 1).
        3. Add a new content unit a new repository version (this create
           repository version 2).
        4. Create a new repository version using version 1 as ``base_version``
           (this creates version 3).
        5. Check that version 1 and version 3 have the same content.
        """
        # create repo version 1
        repo = self.create_sync_repo()
        version_content = []
        version_content.append(
            sorted(
                [remove_created_key(item) for item in get_content(repo)[FILE_CONTENT_NAME]],
                key=lambda item: item["pulp_href"],
            )
        )
        self.assertIsNone(get_versions(repo)[1]["base_version"])

        content = self.content.pop()

        # create repo version 2
        modify_repo(self.cfg, repo, add_units=[content])
        repo = self.client.get(repo["pulp_href"])

        # create repo version 3 from version 1
        base_version = get_versions(repo)[1]["pulp_href"]
        modify_repo(self.cfg, repo, base_version=base_version)
        repo = self.client.get(repo["pulp_href"])

        # assert that base_version of the version 3 points to version 1
        self.assertEqual(get_versions(repo)[3]["base_version"], base_version)

        # assert that content on version 1 is equal to content on version 3
        version_content.append(
            sorted(
                [remove_created_key(item) for item in get_content(repo)[FILE_CONTENT_NAME]],
                key=lambda item: item["pulp_href"],
            )
        )
        self.assertEqual(version_content[0], version_content[1], version_content)

    def test_different_repository(self):
        """Test ``base_version`` for different repositories.

        Do the following:

        1. Create a new repository A and sync it.
        2. Create a new repository B and a new version for this repository
           specify repository A version 1 as the ``base_version``.
        3. Check that repository A version 1 and repository B version 1 have
           the same content.
        """
        # create repo A
        repo = self.create_sync_repo()
        version_content = []
        version_content.append(
            sorted(
                [remove_created_key(item) for item in get_content(repo)[FILE_CONTENT_NAME]],
                key=lambda item: item["pulp_href"],
            )
        )
        self.assertIsNone(get_versions(repo)[1]["base_version"])

        # get repo A version 1 to be used as base_version
        base_version = get_versions(repo)[1]["pulp_href"]

        # create repo B
        repo = self.client.post(FILE_REPO_PATH, gen_repo())
        self.addCleanup(self.client.delete, repo["pulp_href"])

        # create a version for repo B using repo A version 1 as base_version
        modify_repo(self.cfg, repo, base_version=base_version)
        repo = self.client.get(repo["pulp_href"])

        # assert that base_version of repo B points to version 1 of repo A
        self.assertEqual(get_versions(repo)[1]["base_version"], base_version)

        # assert that content on version 1 of repo A is equal to content on
        # version 1 repo B
        version_content.append(
            sorted(
                [remove_created_key(item) for item in get_content(repo)[FILE_CONTENT_NAME]],
                key=lambda item: item["pulp_href"],
            )
        )

        self.assertEqual(version_content[0], version_content[1], version_content)

    def test_base_version_other_parameters(self):
        """Test ``base_version`` can be used together with other parameters.

        ``add_content_units`` and ``remove_content_units``.
        """
        # create repo version 1
        self.skipTest("Temporarily skipping while we figure out a better testing strategy.")
        repo = self.create_sync_repo()
        version_1_content = [
            remove_created_key(item) for item in get_content(repo)[FILE_CONTENT_NAME]
        ]
        self.assertIsNone(get_versions(repo)[1]["base_version"])

        # create repo version 2 from version 1
        base_version = get_versions(repo)[1]["pulp_href"]
        added_content = remove_created_key(self.content.pop())
        removed_content = choice(version_1_content)
        modify_repo(
            self.cfg,
            repo,
            base_version=base_version,
            add_units=[added_content],
            remove_units=[removed_content],
        )
        repo = self.client.get(repo["pulp_href"])
        version_2_content = [
            remove_created_key(item) for item in get_content(repo)[FILE_CONTENT_NAME]
        ]

        # assert that base_version of the version 2 points to version 1
        self.assertEqual(get_versions(repo)[2]["base_version"], base_version)

        # assert that the removed content is not present on repo version 2
        self.assertNotIn(removed_content, version_2_content)

        # assert that the added content is present on repo version 2
        self.assertIn(added_content, version_2_content)

        # assert that the same amount of units are present in both versions
        self.assertEqual(len(version_1_content), len(version_2_content))

    def test_base_version_exception(self):
        """Exception is raised when non-existent ``base_version`` is used.

        Do the following:

        1. Create a repository B and an attempt to specify a non-existent
           ``base_version``.
        3. Assert that an HTTP exception is raised.
        """
        repo = self.client.post(FILE_REPO_PATH, gen_repo())
        self.addCleanup(self.client.delete, repo["pulp_href"])

        with self.assertRaises(HTTPError):
            modify_repo(self.cfg, repo, base_version=utils.uuid4())

    def create_sync_repo(self):
        """Create, and sync a repo."""
        repo = self.client.post(FILE_REPO_PATH, gen_repo())
        self.addCleanup(self.client.delete, repo["pulp_href"])

        body = gen_file_remote(url=FILE_FIXTURE_MANIFEST_URL)
        remote = self.client.post(FILE_REMOTE_PATH, body)
        self.addCleanup(self.client.delete, remote["pulp_href"])

        sync(self.cfg, remote, repo)
        return self.client.get(repo["pulp_href"])


class UpdateRepoVersionTestCase(unittest.TestCase):
    """Repository version can not be updated using PATCH or PUT.

    Assert that an HTTP exception is raised.

    This test targets the following issue:

    * `Pulp #4667 <https://pulp.plan.io/issues/4667>`_
    """

    @classmethod
    def setUpClass(cls):
        """Create class-wide variables."""
        cls.cfg = config.get_config()
        cls.client = api.Client(cls.cfg)

    def test_http_error(self):
        """Test partial update repository version."""
        remote = self.client.post(FILE_REMOTE_PATH, gen_file_remote())
        self.addCleanup(self.client.delete, remote["pulp_href"])

        repo = self.client.post(FILE_REPO_PATH, gen_repo())
        self.addCleanup(self.client.delete, repo["pulp_href"])

        # create repo version
        sync(self.cfg, remote, repo)
        repo = self.client.get(repo["pulp_href"])

        self.assert_patch(repo)
        self.assert_put(repo)

    def assert_patch(self, repo):
        """Assert PATCH method raises an HTTP exception."""
        previous_repo_name = repo["name"]
        with self.assertRaises(HTTPError):
            self.client.patch(repo["latest_version_href"], {"name": utils.uuid4()})
        repo = self.client.get(repo["pulp_href"])
        self.assertEqual(previous_repo_name, repo["name"], repo)

    def assert_put(self, repo):
        """Assert PUT method raises an HTTP exception."""
        previous_repo_name = repo["name"]
        with self.assertRaises(HTTPError):
            repo["name"] = utils.uuid4()
            self.client.put(repo["latest_version_href"], repo)
        repo = self.client.get(repo["pulp_href"])
        self.assertEqual(previous_repo_name, repo["name"], repo)


class FilterArtifactsTestCase(unittest.TestCase):
    """Filter artifacts by repository version.

    This test targets the following issue:

    * `Pulp #4811 <https://pulp.plan.io/issues/4811>`_
    """

    @classmethod
    def setUpClass(cls):
        """Create class-wide variables.

        Populate Pulp with artifacts to show how the filter is related to
        repository version.
        """
        cls.cfg = config.get_config()
        populate_pulp(cls.cfg, url=FILE_MANY_FIXTURE_MANIFEST_URL)
        cls.client = api.Client(cls.cfg)

    def test_filter_last_repository_version(self):
        """Filter by last repository version.

        For a repository with more than one version.
        """
        repo = self.client.post(FILE_REPO_PATH, gen_repo())
        self.addCleanup(self.client.delete, repo["pulp_href"])

        for url in [FILE2_FIXTURE_MANIFEST_URL, FILE_FIXTURE_MANIFEST_URL]:
            remote = self.client.post(FILE_REMOTE_PATH, gen_file_remote(url=url))
            self.addCleanup(self.client.delete, remote["pulp_href"])
            sync(self.cfg, remote, repo)
            repo = self.client.get(repo["pulp_href"])

        artifacts = self.client.get(
            ARTIFACTS_PATH, params={"repository_version": repo["latest_version_href"]}
        )
        # Even though every sync adds 3 content units to the repository the fixture data contains
        # the same relative urls so the second sync replaces the first 3, leaving a total of 3 each
        # time
        self.assertEqual(len(artifacts), FILE_FIXTURE_COUNT, artifacts)

    def test_filter_invalid_repo_version(self):
        """Filter by invalid repository version."""
        repo = self.client.post(FILE_REPO_PATH, gen_repo())
        self.addCleanup(self.client.delete, repo["pulp_href"])
        with self.assertRaises(HTTPError) as ctx:
            self.client.using_handler(api.json_handler).get(
                ARTIFACTS_PATH, params={"repository_version": repo["pulp_href"]}
            )
        for key in ("uri", "repositoryversion", "not", "found"):
            self.assertIn(key, ctx.exception.response.json()[0].lower(), ctx.exception.response)

    def test_filter_valid_repo_version(self):
        """Filter by valid repository version."""
        remote = self.client.post(FILE_REMOTE_PATH, gen_file_remote())
        self.addCleanup(self.client.delete, remote["pulp_href"])
        repo = self.client.post(FILE_REPO_PATH, gen_repo())
        self.addCleanup(self.client.delete, repo["pulp_href"])
        sync(self.cfg, remote, repo)
        repo = self.client.get(repo["pulp_href"])
        artifacts = self.client.get(
            ARTIFACTS_PATH, params={"repository_version": repo["latest_version_href"]}
        )
        self.assertEqual(len(artifacts), FILE_FIXTURE_COUNT, artifacts)


class DeleteRepoVersionResourcesTestCase(unittest.TestCase):
    """Test whether removing a repository version affects related resources.

    Test whether removing a repository version will remove a related Publication.
    Test whether removing a repository version a Distribution will not be removed.

    This test targets the following issue:

    `Pulp #5303 <https://pulp.plan.io/issues/5303>`_
    """

    @classmethod
    def setUpClass(cls):
        """Create class-wide variables."""
        cls.cfg = config.get_config()
        cls.client = api.Client(cls.cfg)

    def test_delete_publication(self):
        """Publication is removed once the repository version is removed."""
        repo = self.create_sync_repo(2)
        version_href = self.client.get(repo["versions_href"])[0]["pulp_href"]
        publication = create_file_publication(self.cfg, repo, version_href)

        # delete repo version used to create publication
        self.client.delete(version_href)

        with self.assertRaises(HTTPError) as ctx:
            self.client.get(publication["pulp_href"])

        for key in ("not", "found"):
            self.assertIn(
                key, ctx.exception.response.json()["detail"].lower(), ctx.exception.response
            )

    def test_delete_distribution(self):
        """Distribution is not removed once repository version is removed."""
        repo = self.create_sync_repo(2)
        version_href = self.client.get(repo["versions_href"])[0]["pulp_href"]
        publication = create_file_publication(self.cfg, repo, version_href)

        distribution = self.client.post(
            FILE_DISTRIBUTION_PATH, gen_distribution(publication=publication["pulp_href"])
        )
        self.addCleanup(self.client.delete, distribution["pulp_href"])

        # delete repo version used to create publication
        self.client.delete(version_href)

        updated_distribution = self.client.get(distribution["pulp_href"])
        self.assertIsNone(updated_distribution["publication"], updated_distribution)

    def create_sync_repo(self, number_syncs=1):
        """Create and sync a repository.

        Given the number of times to be synced.
        """
        repo = self.client.post(FILE_REPO_PATH, gen_repo())
        self.addCleanup(self.client.delete, repo["pulp_href"])

        remote = self.client.post(FILE_REMOTE_PATH, gen_file_remote())
        self.addCleanup(self.client.delete, remote["pulp_href"])

        for _ in range(number_syncs):
            sync(self.cfg, remote, repo)
        return self.client.get(repo["pulp_href"])


class ClearAllUnitsRepoVersionTestCase(unittest.TestCase):
    """Test clear of all units of a given repository version.

    This test targets the following issue:

    `Pulp #4956 <https://pulp.plan.io/issues/4956>`_
    """

    @classmethod
    def setUpClass(cls):
        """Add content to Pulp."""
        cls.cfg = config.get_config()
        cls.client = api.Client(cls.cfg)
        # Populate Pulp to create content units.
        populate_pulp(cls.cfg, url=FILE_LARGE_FIXTURE_MANIFEST_URL)
        cls.content = sample(cls.client.get(FILE_CONTENT_PATH), 10)

    def setUp(self):
        """Create and sync a repository."""
        self.repo = self.client.post(FILE_REPO_PATH, gen_repo())
        self.addCleanup(self.client.delete, self.repo["pulp_href"])
        remote = self.client.post(FILE_REMOTE_PATH, gen_file_remote())
        self.addCleanup(self.client.delete, remote["pulp_href"])
        sync(self.cfg, remote, self.repo)
        self.repo = self.client.get(self.repo["pulp_href"])

    def test_add_and_clear_all_units(self):
        """Test addition and removal of all units for a given repository version."""
        content = choice(self.content)
        modify_repo(self.cfg, self.repo, add_units=[content], remove_units=["*"])
        self.repo = self.client.get(self.repo["pulp_href"])

        added_content = get_content(self.repo)[FILE_CONTENT_NAME]
        self.assertEqual(len(added_content), 1, added_content)

        self.assertEqual(remove_created_key(content), remove_created_key(added_content[0]))

    def test_clear_all_units_using_base_version(self):
        """Test clear all units using base version."""
        for content in self.content:
            modify_repo(self.cfg, self.repo, add_units=[content])

        self.repo = self.client.get(self.repo["pulp_href"])
        base_version = get_versions(self.repo)[0]["pulp_href"]

        modify_repo(self.cfg, self.repo, base_version=base_version, remove_units=["*"])
        self.repo = self.client.get(self.repo["pulp_href"])

        content_last_version = get_content(self.repo)[FILE_CONTENT_NAME]
        self.assertEqual(len(content_last_version), 0, content_last_version)

    def test_clear_all_units(self):
        """Test clear all units of a given repository version."""
        added_content = sorted(
            [content["pulp_href"] for content in get_content(self.repo)[FILE_CONTENT_NAME]]
        )

        modify_repo(self.cfg, self.repo, remove_units=["*"])
        self.repo = self.client.get(self.repo["pulp_href"])
        removed_content = sorted(
            [content["pulp_href"] for content in get_removed_content(self.repo)[FILE_CONTENT_NAME]]
        )
        self.assertEqual(added_content, removed_content)
        content = get_content(self.repo)[FILE_CONTENT_NAME]
        self.assertEqual(len(content), 0, content)

    def test_http_error(self):
        """Test http error is raised."""
        added_content = choice(get_added_content(self.repo)[FILE_CONTENT_NAME])
        with self.assertRaises(HTTPError) as ctx:
            self.client.post(
                self.repo["pulp_href"] + "modify/",
                {"remove_content_units": ["*", added_content["pulp_href"]]},
            )
        for key in ("content", "units", "*"):
            self.assertIn(
                key,
                ctx.exception.response.json()["remove_content_units"][0].lower(),
                ctx.exception.response,
            )


class BaseVersionTestCase(unittest.TestCase):
    """Associate different Content units with the same ``relative_path`` in one RepositoryVersion.

    This test targets the following issues:

    *  `Pulp #4028 <https://pulp.plan.io/issue/4028>`_
    """

    @classmethod
    def setUpClass(cls):
        """Create class-wide variables."""
        cls.cfg = config.get_config()
        cls.client = api.Client(cls.cfg, api.json_handler)

    @classmethod
    def tearDownClass(cls):
        """Clean created resources."""
        delete_orphans()

    def test_add_content_with_base_version(self):
        """Test modify repository with base_version"""
        delete_orphans()

        repo = self.client.post(FILE_REPO_PATH, gen_repo())
        self.addCleanup(self.client.delete, repo["pulp_href"])

        files = {"file": utils.http_get(FILE_URL)}
        artifact = self.client.post(ARTIFACTS_PATH, files=files)

        # create first content unit.
        content_attrs = {"artifact": artifact["pulp_href"], "relative_path": utils.uuid4()}
        content = self.client.using_handler(api.task_handler).post(FILE_CONTENT_PATH, content_attrs)
        repo_version = modify_repo(self.cfg, repo, add_units=[content])
        repo = self.client.get(repo["pulp_href"])

        self.assertEqual(get_content(repo)[FILE_CONTENT_NAME][0], content)

        files = {"file": utils.http_get(FILE2_URL)}
        artifact = self.client.post(ARTIFACTS_PATH, files=files)

        # create second content unit.
        second_content_attrs = {
            "artifact": artifact["pulp_href"],
            "relative_path": content_attrs["relative_path"],
        }
        content2 = self.client.using_handler(api.task_handler).post(
            FILE_CONTENT_PATH, second_content_attrs
        )
        modify_repo(self.cfg, repo, add_units=[content2])
        repo = self.client.get(repo["pulp_href"])

        self.assertEqual(get_content(repo)[FILE_CONTENT_NAME][0], content2)

        modify_repo(self.cfg, repo, base_version=repo_version["pulp_href"], add_units=[content2])
        repo = self.client.get(repo["pulp_href"])

        self.assertEqual(get_content(repo)[FILE_CONTENT_NAME][0], content2)


class RepoVersionRetentionTestCase(unittest.TestCase):
    """Test retain_repo_versions for repositories

    This test targets the following issues:

    *  `Pulp #8368 <https:://pulp.plan.io/issues/8368>`_
    """

    @classmethod
    def setUp(self):
        """Add content to Pulp."""
        self.cfg = config.get_config()
        self.client = api.Client(self.cfg, api.json_handler)
        self.core_client = CoreApiClient(configuration=self.cfg.get_bindings_config())
        self.file_client = gen_file_client()

        self.content_api = ContentFilesApi(self.file_client)
        self.repo_api = RepositoriesFileApi(self.file_client)
        self.version_api = RepositoriesFileVersionsApi(self.file_client)
        self.distro_api = DistributionsFileApi(self.file_client)
        self.publication_api = PublicationsFileApi(self.file_client)

        delete_orphans()
        populate_pulp(self.cfg, url=FILE_LARGE_FIXTURE_MANIFEST_URL)
        self.content = sample(self.content_api.list().results, 3)
        self.publications = []

    def _create_repo_versions(self, repo_attributes={}):
        self.repo = self.repo_api.create(gen_repo(**repo_attributes))
        self.addCleanup(self.repo_api.delete, self.repo.pulp_href)

        if "autopublish" in repo_attributes and repo_attributes["autopublish"]:
            self.distro = create_distribution(repository_href=self.repo.pulp_href)
            self.addCleanup(self.distro_api.delete, self.distro.pulp_href)

        for content in self.content:
            result = self.repo_api.modify(
                self.repo.pulp_href, {"add_content_units": [content.pulp_href]}
            )
            monitor_task(result.task)
            self.repo = self.repo_api.read(self.repo.pulp_href)
            self.publications += self.publication_api.list(
                repository_version=self.repo.latest_version_href
            ).results

    def test_retain_repo_versions(self):
        """Test repo version retention."""
        self._create_repo_versions({"retain_repo_versions": 1})

        versions = self.version_api.list(file_file_repository_href=self.repo.pulp_href).results
        self.assertEqual(len(versions), 1)

        latest_version = self.version_api.read(
            file_file_repository_version_href=self.repo.latest_version_href
        )
        self.assertEqual(latest_version.number, 3)
        self.assertEqual(latest_version.content_summary.present["file.file"]["count"], 3)
        self.assertEqual(latest_version.content_summary.added["file.file"]["count"], 3)

    def test_retain_repo_versions_on_update(self):
        """Test repo version retention when retain_repo_versions is set."""
        self._create_repo_versions()

        versions = self.version_api.list(file_file_repository_href=self.repo.pulp_href).results
        self.assertEqual(len(versions), 4)

        # update retain_repo_versions to 2
        result = self.repo_api.partial_update(self.repo.pulp_href, {"retain_repo_versions": 2})
        monitor_task(result.task)

        versions = self.version_api.list(file_file_repository_href=self.repo.pulp_href).results
        self.assertEqual(len(versions), 2)

        latest_version = self.version_api.read(
            file_file_repository_version_href=self.repo.latest_version_href
        )
        self.assertEqual(latest_version.number, 3)
        self.assertEqual(latest_version.content_summary.present["file.file"]["count"], 3)
        self.assertEqual(latest_version.content_summary.added["file.file"]["count"], 1)

    def test_autodistribute(self):
        """Test repo version retention with autopublish/autodistribute."""
        self._create_repo_versions({"retain_repo_versions": 1, "autopublish": True})

        # all but the last publication should be gone
        for publication in self.publications[:-1]:
            with self.assertRaises(ApiException) as ae:
                self.publication_api.read(publication.pulp_href)
            self.assertEqual(404, ae.exception.status)

        # check that the last publication is distributed
        manifest = download_content_unit(self.cfg, self.distro.to_dict(), "PULP_MANIFEST")
        self.assertEqual(manifest.decode("utf-8").count("\n"), len(self.content))


class ContentInRepositoryVersionViewTestCase(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        """Create class-wide variables."""
        cls.cfg = config.get_config()
        cls.file_client = gen_file_client()
        cls.remote_api = RemotesFileApi(cls.file_client)
        cls.repo_api = RepositoriesFileApi(cls.file_client)
        cls.repo_ver_api = RepositoryVersionsApi(cls.file_client)

    @classmethod
    def tearDownClass(cls):
        delete_orphans()

    def test_all(self):
        """Sync two repositories and check view filter."""
        # Test content doesn't exists.
        non_existant_content_href = (
            "/pulp/api/v3/content/file/files/c4ed74cf-a806-490d-a25f-94c3c3dd2dd7/"
        )

        with self.assertRaises(ApiException) as ctx:
            self.repo_ver_api.list(content=non_existant_content_href)

        self.assertEqual(ctx.exception.status, 400)

        initial_rv_count = self.repo_ver_api.list(limit=1).count

        repo = self.repo_api.create(gen_repo())
        self.addCleanup(self.repo_api.delete, repo.pulp_href)

        repo_second = self.repo_api.create(gen_repo())
        self.addCleanup(self.repo_api.delete, repo_second.pulp_href)

        remote = self.remote_api.create(gen_file_remote())
        self.addCleanup(self.remote_api.delete, remote.pulp_href)

        body = gen_file_remote(url=FILE2_FIXTURE_MANIFEST_URL)
        remote_second = self.remote_api.create(body)
        self.addCleanup(self.remote_api.delete, remote_second.pulp_href)

        repo_sync_data = RepositorySyncURL(remote=remote.pulp_href)
        repo_sync_data_second = RepositorySyncURL(remote=remote_second.pulp_href)

        sync_response = self.repo_api.sync(repo.pulp_href, repo_sync_data)
        monitor_task(sync_response.task)

        sync_response_second = self.repo_api.sync(repo_second.pulp_href, repo_sync_data_second)
        monitor_task(sync_response_second.task)

        # Update repository data and get one content unit from first repository.
        repo = self.repo_api.read(repo.pulp_href)
        content_href = get_content(repo.to_dict())[FILE_CONTENT_NAME][0]["pulp_href"]

        rv_total = len(self.repo_ver_api.list().to_dict()["results"])
        rv_search = self.repo_ver_api.list(content=content_href).to_dict()["results"]

        # Test only one repostiory version has selected content.
        self.assertEqual(len(rv_search), 1)
        # Test if repositories version with content matches.
        self.assertEqual(rv_search[0]["pulp_href"], repo.latest_version_href)
        # Test total number of repository version. Two for each repository.
        self.assertEqual(rv_total - initial_rv_count, 4)



# Generated by Django 2.2.11 on 2020-04-07 15:47

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0026_task_group'),
    ]

    operations = [
        migrations.CreateModel(
            name='PulpExport',
            fields=[
                ('export_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, related_name='core_pulp_export', serialize=False, to='core.Export')),
                ('sha256', models.CharField(max_length=64, null=True)),
                ('filename', models.CharField(max_length=4096, null=True)),
            ],
            options={
                'default_related_name': '%(app_label)s_pulp_export',
            },
            bases=('core.export',),
        ),
        migrations.AlterField(
            model_name='pulpexporter',
            name='last_export',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.PROTECT, related_name='last_export', to='core.PulpExport'),
        ),
    ]



# Generated by Django 2.2.17 on 2020-11-12 19:38

import django.core.validators
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0050_namespace_access_policies'),
    ]

    operations = [
        migrations.AddField(
            model_name='remote',
            name='connect_timeout',
            field=models.FloatField(null=True, validators=[django.core.validators.MinValueValidator(0.0, 'Timeout must be >= 0')]),
        ),
        migrations.AddField(
            model_name='remote',
            name='sock_connect_timeout',
            field=models.FloatField(null=True, validators=[django.core.validators.MinValueValidator(0.0, 'Timeout must be >= 0')]),
        ),
        migrations.AddField(
            model_name='remote',
            name='sock_read_timeout',
            field=models.FloatField(null=True, validators=[django.core.validators.MinValueValidator(0.0, 'Timeout must be >= 0')]),
        ),
        migrations.AddField(
            model_name='remote',
            name='total_timeout',
            field=models.FloatField(null=True, validators=[django.core.validators.MinValueValidator(0.0, 'Timeout must be >= 0')]),
        ),
    ]



import asyncio
import logging
import mimetypes
import os
import re
from gettext import gettext as _

from aiohttp.client_exceptions import ClientResponseError
from aiohttp.web import FileResponse, StreamResponse, HTTPOk
from aiohttp.web_exceptions import HTTPForbidden, HTTPFound, HTTPNotFound
from yarl import URL

from asgiref.sync import sync_to_async

import django

os.environ.setdefault("DJANGO_SETTINGS_MODULE", "pulpcore.app.settings")
django.setup()

from django.conf import settings  # noqa: E402: module level not at top of file
from django.core.exceptions import (  # noqa: E402: module level not at top of file
    MultipleObjectsReturned,
    ObjectDoesNotExist,
)
from django.db import (  # noqa: E402: module level not at top of file
    connection,
    DatabaseError,
    IntegrityError,
    transaction,
)
from pulpcore.app.models import (  # noqa: E402: module level not at top of file
    Artifact,
    ContentArtifact,
    Distribution,
    Publication,
    Remote,
    RemoteArtifact,
)
from pulpcore.exceptions import UnsupportedDigestValidationError  # noqa: E402

from jinja2 import Template  # noqa: E402: module level not at top of file
from pulpcore.cache import ContentCache  # noqa: E402

log = logging.getLogger(__name__)


class PathNotResolved(HTTPNotFound):
    """
    The path could not be resolved to a published file.

    This could be caused by either the distribution, the publication,
    or the published file could not be found.
    """

    def __init__(self, path, *args, **kwargs):
        """Initialize the Exception."""
        self.path = path
        super().__init__(*args, **kwargs)


class ArtifactNotFound(Exception):
    """
    The artifact associated with a published-artifact does not exist.
    """

    pass


class Handler:
    """
    A default Handler for the Content App that also can be subclassed to create custom handlers.

    This Handler will perform the following:

    1. Match the request against a Distribution

    2. Call the certguard check if a certguard exists for the matched Distribution.

    3. If the Distribution has a `publication` serve that Publication's `PublishedArtifacts`,
       `PublishedMetadata` by the remaining `relative path`. If still unserved and if `pass_through`
       is set, the associated `repository_version` will have its `ContentArtifacts` served by
       `relative_path` also. This will serve the associated `Artifact`.

    4. If still unmatched, and the Distribution has a `repository` attribute set, find it's latest
       `repository_version`. If the Distribution has a `repository_version` attribute set, use that.
       For this `repository_version`, find matching `ContentArtifact` objects by `relative_path` and
       serve them. If there is an associated `Artifact` return it.

    5. If the Distribution has a `remote`, find an associated `RemoteArtifact` that matches by
       `relative_path`. Fetch and stream the corresponding `RemoteArtifact` to the client,
       optionally saving the `Artifact` depending on the `policy` attribute.

    """

    hop_by_hop_headers = [
        "connection",
        "content-encoding",
        "content-length",
        "keep-alive",
        "public",
        "proxy-authenticate",
        "transfer-encoding",
        "upgrade",
    ]

    distribution_model = None

    @staticmethod
    def _reset_db_connection():
        """
        Reset database connection if it's unusable or obselete to avoid "connection already closed".
        """
        connection.close_if_unusable_or_obsolete()

    async def list_distributions(self, request):
        """
        The handler for an HTML listing all distributions

        Args:
            request (:class:`aiohttp.web.request`): The request from the client.

        Returns:
            :class:`aiohttp.web.HTTPOk`: The response back to the client.
        """

        def get_base_paths_blocking():
            if self.distribution_model is None:
                base_paths = list(Distribution.objects.values_list("base_path", flat=True))
            else:
                base_paths = list(
                    self.distribution_model.objects.values_list("base_path", flat=True)
                )
            return base_paths

        base_paths = await sync_to_async(get_base_paths_blocking)()
        directory_list = ["{}/".format(path) for path in base_paths]
        return HTTPOk(headers={"Content-Type": "text/html"}, body=self.render_html(directory_list))

    @classmethod
    async def find_base_path_cached(cls, request, cached):
        """
        Finds the base-path to use for the base-key in the cache

        Args:
            request (:class:`aiohttp.web.request`): The request from the client.
            cached (:class:`CacheAiohttp`): The Pulp cache

        Returns:
            str: The base-path associated with this request
        """
        path = request.match_info["path"]
        base_paths = cls._base_paths(path)
        multiplied_base_paths = []
        for i, base_path in enumerate(base_paths):
            copied_by_index_base_path = [base_path for _ in range(i + 1)]
            multiplied_base_paths.extend(copied_by_index_base_path)
        index_p1 = await cached.exists(base_key=multiplied_base_paths)
        if index_p1:
            return base_paths[index_p1 - 1]
        else:
            distro = await sync_to_async(cls._match_distribution)(path)
            return distro.base_path

    @classmethod
    async def auth_cached(cls, request, cached, base_key):
        """
        Authentication check for the cached stream_content handler

        Args:
            request (:class:`aiohttp.web.request`): The request from the client.
            cached (:class:`CacheAiohttp`): The Pulp cache
            base_key (str): The base_key associated with this response
        """
        guard_key = "DISTRO#GUARD#PRESENT"
        present = await cached.get(guard_key, base_key=base_key)
        if present == b"True" or present is None:
            path = request.match_info["path"]
            distro = await sync_to_async(cls._match_distribution)(path)
            try:
                guard = await sync_to_async(cls._permit)(request, distro)
            except HTTPForbidden:
                guard = True
                raise
            finally:
                if not present:
                    await cached.set(guard_key, str(guard), base_key=base_key)

    @ContentCache(
        base_key=lambda req, cac: Handler.find_base_path_cached(req, cac),
        auth=lambda req, cac, bk: Handler.auth_cached(req, cac, bk),
    )
    async def stream_content(self, request):
        """
        The request handler for the Content app.

        Args:
            request (:class:`aiohttp.web.request`): The request from the client.

        Returns:
            :class:`aiohttp.web.StreamResponse` or :class:`aiohttp.web.FileResponse`: The response
                back to the client.
        """
        path = request.match_info["path"]
        return await self._match_and_stream(path, request)

    @staticmethod
    def _base_paths(path):
        """
        Get a list of base paths used to match a distribution.

        Args:
            path (str): The path component of the URL.

        Returns:
            list: Of base paths.

        """
        tree = []
        while True:
            base = os.path.split(path)[0]
            if not base.lstrip("/"):
                break
            tree.append(base)
            path = base
        return tree

    @classmethod
    def _match_distribution(cls, path):
        """
        Match a distribution using a list of base paths and return its detail object.

        Args:
            path (str): The path component of the URL.

        Returns:
            The detail object of the matched distribution.

        Raises:
            PathNotResolved: when not matched.
        """
        base_paths = cls._base_paths(path)
        if cls.distribution_model is None:
            try:
                return (
                    Distribution.objects.select_related(
                        "repository", "repository_version", "publication", "remote"
                    )
                    .get(base_path__in=base_paths)
                    .cast()
                )
            except ObjectDoesNotExist:
                log.debug(
                    _("Distribution not matched for {path} using: {base_paths}").format(
                        path=path, base_paths=base_paths
                    )
                )
        else:
            try:
                return cls.distribution_model.objects.select_related(
                    "repository", "repository_version", "publication", "remote"
                ).get(base_path__in=base_paths)
            except ObjectDoesNotExist:
                log.debug(
                    _("Distribution not matched for {path} using: {base_paths}").format(
                        path=path, base_paths=base_paths
                    )
                )
        raise PathNotResolved(path)

    @staticmethod
    def _permit(request, distribution):
        """
        Permit the request.

        Authorization is delegated to the optional content-guard associated with the distribution.

        Args:
            request (:class:`aiohttp.web.Request`): A request for a published file.
            distribution (detail of :class:`pulpcore.plugin.models.Distribution`): The matched
                distribution.

        Raises:
            :class:`aiohttp.web_exceptions.HTTPForbidden`: When not permitted.
        """
        guard = distribution.content_guard
        if not guard:
            return False
        try:
            guard.cast().permit(request)
        except PermissionError as pe:
            log.debug(
                _('Path: %(p)s not permitted by guard: "%(g)s" reason: %(r)s'),
                {"p": request.path, "g": guard.name, "r": str(pe)},
            )
            raise HTTPForbidden(reason=str(pe))
        return True

    @staticmethod
    def response_headers(path):
        """
        Get the Content-Type and Encoding-Type headers for the requested `path`.

        Args:
            path (str): The relative path that was requested.

        Returns:
            headers (dict): A dictionary of response headers.
        """
        content_type, encoding = mimetypes.guess_type(path)
        headers = {}
        if content_type:
            headers["Content-Type"] = content_type
        return headers

    @staticmethod
    def render_html(directory_list):
        """
        Render a list of strings as an HTML list of links.

        Args:
            directory_list (iterable): an iterable of strings representing file and directory names

        Returns:
            String representing HTML of the directory listing.
        """
        template = Template(
            """
        <!DOCTYPE html>
        <html>
            <body>
                <ul>
                {% for name in dir_list %}
                    <li><a href="{{ name|e }}">{{ name|e }}</a></li>
                {% endfor %}
                </ul>
            </body>
        </html>
        """
        )
        return template.render(dir_list=sorted(directory_list))

    async def list_directory(self, repo_version, publication, path):
        """
        Generate a set with directory listing of the path.

        This method expects either a repo_version or a publication in addition to a path. This
        method generates a set of strings representing the list of a path inside the repository
        version or publication.

        Args:
            repo_version (:class:`~pulpcore.app.models.RepositoryVersion`): The repository version
            publication (:class:`~pulpcore.app.models.Publication`): Publication
            path (str): relative path inside the repo version of publication.

        Returns:
            Set of strings representing the files and directories in the directory listing.
        """

        def file_or_directory_name(directory_path, relative_path):
            result = re.match(r"({})([^\/]*)(\/*)".format(directory_path), relative_path)
            return "{}{}".format(result.groups()[1], result.groups()[2])

        def list_directory_blocking():
            if not publication and not repo_version:
                raise Exception("Either a repo_version or publication is required.")
            if publication and repo_version:
                raise Exception("Either a repo_version or publication can be specified.")

            directory_list = set()

            if publication:
                pas = publication.published_artifact.filter(relative_path__startswith=path)
                for pa in pas:
                    directory_list.add(file_or_directory_name(path, pa.relative_path))

                if publication.pass_through:
                    cas = ContentArtifact.objects.filter(
                        content__in=publication.repository_version.content,
                        relative_path__startswith=path,
                    )
                    for ca in cas:
                        directory_list.add(file_or_directory_name(path, ca.relative_path))

            if repo_version:
                cas = ContentArtifact.objects.filter(
                    content__in=repo_version.content, relative_path__startswith=path
                )
                for ca in cas:
                    directory_list.add(file_or_directory_name(path, ca.relative_path))

            if directory_list:
                return directory_list
            else:
                raise PathNotResolved(path)

        return await sync_to_async(list_directory_blocking)()

    async def _match_and_stream(self, path, request):
        """
        Match the path and stream results either from the filesystem or by downloading new data.

        After deciding the client can access the distribution at ``path``, this function calls
        :meth:`Distribution.content_handler`. If that function returns a not-None result, it is
        returned to the client.

        Then the publication linked to the Distribution is used to determine what content should
        be served. If ``path`` is a directory entry (i.e. not a file), the directory contents
        are served to the client. This method calls
        :meth:`Distribution.content_handler_list_directory` to acquire any additional entries the
        Distribution's content_handler might serve in that directory. If there is an Artifact to be
        served, it is served to the client.

        If there's no publication, the above paragraph is applied to the latest repository linked
        to the matched Distribution.

        Finally, when nothing is served to client yet, we check if there is a remote for the
        Distribution. If so, the Artifact is pulled from the remote and streamed to the client.

        Args:
            path (str): The path component of the URL.
            request(:class:`~aiohttp.web.Request`): The request to prepare a response for.

        Raises:
            PathNotResolved: The path could not be matched to a published file.
            PermissionError: When not permitted.

        Returns:
            :class:`aiohttp.web.StreamResponse` or :class:`aiohttp.web.FileResponse`: The response
                streamed back to the client.
        """

        distro = await sync_to_async(self._match_distribution)(path)

        await sync_to_async(self._permit)(request, distro)

        rel_path = path.lstrip("/")
        rel_path = rel_path[len(distro.base_path) :]
        rel_path = rel_path.lstrip("/")

        content_handler_result = await sync_to_async(distro.content_handler)(rel_path)
        if content_handler_result is not None:
            return content_handler_result

        headers = self.response_headers(rel_path)

        repository = distro.repository
        publication = distro.publication
        repo_version = distro.repository_version

        if repository:

            def get_latest_publication_or_version_blocking():
                nonlocal repo_version
                nonlocal publication

                # Search for publication serving the closest latest version
                if not publication:
                    try:
                        versions = repository.versions.all()
                        publications = Publication.objects.filter(
                            repository_version__in=versions, complete=True
                        )
                        publication = publications.select_related("repository_version").latest(
                            "repository_version", "pulp_created"
                        )
                        repo_version = publication.repository_version
                    except ObjectDoesNotExist:
                        pass

                if not repo_version:
                    repo_version = repository.latest_version()

            await sync_to_async(get_latest_publication_or_version_blocking)()

        if publication:
            if rel_path == "" or rel_path[-1] == "/":
                try:
                    index_path = "{}index.html".format(rel_path)

                    await sync_to_async(publication.published_artifact.get)(
                        relative_path=index_path
                    )

                    rel_path = index_path
                    headers = self.response_headers(rel_path)
                except ObjectDoesNotExist:
                    dir_list = await self.list_directory(None, publication, rel_path)
                    dir_list.update(
                        await sync_to_async(distro.content_handler_list_directory)(rel_path)
                    )
                    return HTTPOk(
                        headers={"Content-Type": "text/html"}, body=self.render_html(dir_list)
                    )

            # published artifact
            try:

                def get_contentartifact_blocking():
                    return (
                        publication.published_artifact.select_related(
                            "content_artifact",
                            "content_artifact__artifact",
                        )
                        .get(relative_path=rel_path)
                        .content_artifact
                    )

                ca = await sync_to_async(get_contentartifact_blocking)()
            except ObjectDoesNotExist:
                pass
            else:
                if ca.artifact:
                    return await self._serve_content_artifact(ca, headers, request)
                else:
                    return await self._stream_content_artifact(
                        request, StreamResponse(headers=headers), ca
                    )

            # pass-through
            if publication.pass_through:
                try:

                    def get_contentartifact_blocking():
                        ca = ContentArtifact.objects.select_related("artifact").get(
                            content__in=publication.repository_version.content,
                            relative_path=rel_path,
                        )
                        return ca

                    ca = await sync_to_async(get_contentartifact_blocking)()
                except MultipleObjectsReturned:
                    log.error(
                        _("Multiple (pass-through) matches for {b}/{p}"),
                        {"b": distro.base_path, "p": rel_path},
                    )
                    raise
                except ObjectDoesNotExist:
                    pass
                else:
                    if ca.artifact:
                        return await self._serve_content_artifact(ca, headers, request)
                    else:
                        return await self._stream_content_artifact(
                            request, StreamResponse(headers=headers), ca
                        )

        if repo_version and not publication and not distro.SERVE_FROM_PUBLICATION:
            if rel_path == "" or rel_path[-1] == "/":
                index_path = "{}index.html".format(rel_path)

                def contentartifact_exists_blocking():
                    return ContentArtifact.objects.filter(
                        content__in=repo_version.content, relative_path=index_path
                    ).exists()

                contentartifact_exists = await sync_to_async(contentartifact_exists_blocking)()
                if contentartifact_exists:
                    rel_path = index_path
                else:
                    dir_list = await self.list_directory(repo_version, None, rel_path)
                    dir_list.update(
                        await sync_to_async(distro.content_handler_list_directory)(rel_path)
                    )
                    return HTTPOk(
                        headers={"Content-Type": "text/html"}, body=self.render_html(dir_list)
                    )

            try:

                def get_contentartifact_blocking():
                    ca = ContentArtifact.objects.select_related("artifact").get(
                        content__in=repo_version.content, relative_path=rel_path
                    )
                    return ca

                ca = await sync_to_async(get_contentartifact_blocking)()
            except MultipleObjectsReturned:
                log.error(
                    _("Multiple (pass-through) matches for {b}/{p}"),
                    {"b": distro.base_path, "p": rel_path},
                )
                raise
            except ObjectDoesNotExist:
                pass
            else:
                if ca.artifact:
                    return await self._serve_content_artifact(ca, headers, request)
                else:
                    return await self._stream_content_artifact(
                        request, StreamResponse(headers=headers), ca
                    )

        if distro.remote:

            def cast_remote_blocking():
                return distro.remote.cast()

            remote = await sync_to_async(cast_remote_blocking)()

            try:
                url = remote.get_remote_artifact_url(rel_path, request=request)

                def get_remote_artifact_blocking():
                    ra = RemoteArtifact.objects.select_related(
                        "content_artifact",
                        "content_artifact__artifact",
                        "remote",
                    ).get(remote=remote, url=url)
                    return ra

                ra = await sync_to_async(get_remote_artifact_blocking)()
                ca = ra.content_artifact
                if ca.artifact:
                    return await self._serve_content_artifact(ca, headers, request)
                else:
                    return await self._stream_content_artifact(
                        request, StreamResponse(headers=headers), ca
                    )
            except ObjectDoesNotExist:
                ca = ContentArtifact(relative_path=rel_path)
                ra = RemoteArtifact(remote=remote, url=url, content_artifact=ca)
                return await self._stream_remote_artifact(
                    request, StreamResponse(headers=headers), ra
                )

        raise PathNotResolved(path)

    async def _stream_content_artifact(self, request, response, content_artifact):
        """
        Stream and optionally save a ContentArtifact by requesting it using the associated remote.

        If a fatal download failure occurs while downloading and there are additional
        :class:`~pulpcore.plugin.models.RemoteArtifact` objects associated with the
        :class:`~pulpcore.plugin.models.ContentArtifact` they will also be tried. If all
        :class:`~pulpcore.plugin.models.RemoteArtifact` downloads raise exceptions, an HTTP 502
        error is returned to the client.

        Args:
            request(:class:`~aiohttp.web.Request`): The request to prepare a response for.
            response (:class:`~aiohttp.web.StreamResponse`): The response to stream data to.
            content_artifact (:class:`~pulpcore.plugin.models.ContentArtifact`): The ContentArtifact
                to fetch and then stream back to the client

        Raises:
            :class:`~aiohttp.web.HTTPNotFound` when no
                :class:`~pulpcore.plugin.models.RemoteArtifact` objects associated with the
                :class:`~pulpcore.plugin.models.ContentArtifact` returned the binary data needed for
                the client.
        """

        def get_remote_artifacts_blocking():
            return list(content_artifact.remoteartifact_set.select_related("remote").order_by_acs())

        remote_artifacts = await sync_to_async(get_remote_artifacts_blocking)()
        for remote_artifact in remote_artifacts:
            try:
                response = await self._stream_remote_artifact(request, response, remote_artifact)
                return response

            except (ClientResponseError, UnsupportedDigestValidationError) as e:
                log.warning(
                    _("Could not download remote artifact at '{}': {}").format(
                        remote_artifact.url, str(e)
                    )
                )
                continue

        raise HTTPNotFound()

    def _save_artifact(self, download_result, remote_artifact):
        """
        Create/Get an Artifact and associate it to a RemoteArtifact and/or ContentArtifact.

        Create (or get if already existing) an :class:`~pulpcore.plugin.models.Artifact`
        based on the `download_result` and associate it to the `content_artifact` of the given
        `remote_artifact`. Both the created artifact and the updated content_artifact are saved to
        the DB.  The `remote_artifact` is also saved for the pull-through caching use case.

        Plugin-writers may overide this method if their content module requires
        additional/different steps for saving.

        Args:
            download_result (:class:`~pulpcore.plugin.download.DownloadResult`: The
                DownloadResult for the downloaded artifact.

            remote_artifact (:class:`~pulpcore.plugin.models.RemoteArtifact`): The
                RemoteArtifact to associate the Artifact with.

        Returns:
            The associated :class:`~pulpcore.plugin.models.Artifact`.
        """
        content_artifact = remote_artifact.content_artifact
        remote = remote_artifact.remote
        artifact = Artifact(**download_result.artifact_attributes, file=download_result.path)
        with transaction.atomic():
            try:
                with transaction.atomic():
                    artifact.save()
            except IntegrityError:
                try:
                    artifact = Artifact.objects.get(artifact.q())
                    artifact.touch()
                except (Artifact.DoesNotExist, DatabaseError):
                    # it's possible that orphan cleanup deleted the artifact
                    # so fall back to creating a new artifact again
                    artifact = Artifact(
                        **download_result.artifact_attributes, file=download_result.path
                    )
                    artifact.save()
                else:
                    # The file needs to be unlinked because it was not used to create an artifact.
                    # The artifact must have already been saved while servicing another request for
                    # the same artifact.
                    os.unlink(download_result.path)

            update_content_artifact = True
            if content_artifact._state.adding:
                # This is the first time pull-through content was requested.
                rel_path = content_artifact.relative_path
                c_type = remote.get_remote_artifact_content_type(rel_path)
                content = c_type.init_from_artifact_and_relative_path(artifact, rel_path)
                try:
                    with transaction.atomic():
                        content.save()
                        content_artifact.content = content
                        content_artifact.save()
                except IntegrityError:
                    # There is already content for this Artifact
                    content = c_type.objects.get(content.q())
                    artifacts = content._artifacts
                    if artifact.sha256 != artifacts[0].sha256:
                        raise RuntimeError(
                            "The Artifact downloaded during pull-through does not "
                            "match the Artifact already stored for the same "
                            "content."
                        )
                    content_artifact = ContentArtifact.objects.get(content=content)
                    update_content_artifact = False
                try:
                    with transaction.atomic():
                        remote_artifact.content_artifact = content_artifact
                        remote_artifact.save()
                except IntegrityError:
                    # Remote artifact must have already gotten saved during a parallel request
                    log.info("RemoteArtifact already exists.")
            if update_content_artifact:
                content_artifact.artifact = artifact
                content_artifact.save()
        return artifact

    async def _serve_content_artifact(self, content_artifact, headers, request):
        """
        Handle response for a Content Artifact with the file present.

        Depending on where the file storage (e.g. filesystem, S3, etc) this could be responding with
        the file (filesystem) or a redirect (S3).

        Args:
            content_artifact (:class:`pulpcore.app.models.ContentArtifact`): The Content Artifact to
                respond with.
            headers (dict): A dictionary of response headers.
            request(:class:`~aiohttp.web.Request`): The request to prepare a response for.

        Raises:
            :class:`aiohttp.web_exceptions.HTTPFound`: When we need to redirect to the file
            NotImplementedError: If file is stored in a file storage we can't handle

        Returns:
            The :class:`aiohttp.web.FileResponse` for the file.
        """
        artifact_file = content_artifact.artifact.file
        artifact_name = artifact_file.name

        if settings.DEFAULT_FILE_STORAGE == "pulpcore.app.models.storage.FileSystem":
            return FileResponse(os.path.join(settings.MEDIA_ROOT, artifact_name), headers=headers)
        elif settings.DEFAULT_FILE_STORAGE == "storages.backends.s3boto3.S3Boto3Storage":
            content_disposition = f"attachment;filename={content_artifact.relative_path}"
            parameters = {"ResponseContentDisposition": content_disposition}
            if headers.get("Content-Type"):
                parameters["ResponseContentType"] = headers.get("Content-Type")
            url = URL(
                artifact_file.storage.url(
                    artifact_file.name, parameters=parameters, http_method=request.method
                ),
                encoded=True,
            )
            raise HTTPFound(url)
        elif settings.DEFAULT_FILE_STORAGE == "storages.backends.azure_storage.AzureStorage":
            content_disposition = f"attachment;filename={artifact_name}"
            parameters = {"content_disposition": content_disposition}
            if headers.get("Content-Type"):
                parameters["content_type"] = headers.get("Content-Type")
            url = URL(artifact_file.storage.url(artifact_name, parameters=parameters), encoded=True)
            raise HTTPFound(url)
        else:
            raise NotImplementedError()

    async def _stream_remote_artifact(self, request, response, remote_artifact):
        """
        Stream and save a RemoteArtifact.

        Args:
            request(:class:`~aiohttp.web.Request`): The request to prepare a response for.
            response (:class:`~aiohttp.web.StreamResponse`): The response to stream data to.
            remote_artifact (:class:`~pulpcore.plugin.models.RemoteArtifact`): The RemoteArtifact
                to fetch and then stream back to the client

        Raises:
            :class:`~aiohttp.web.HTTPNotFound` when no
                :class:`~pulpcore.plugin.models.RemoteArtifact` objects associated with the
                :class:`~pulpcore.plugin.models.ContentArtifact` returned the binary data needed for
                the client.

        """

        remote = await sync_to_async(remote_artifact.remote.cast)()

        range_start, range_stop = request.http_range.start, request.http_range.stop
        if range_start or range_stop:
            response.set_status(206)

        async def handle_response_headers(headers):
            for name, value in headers.items():
                lower_name = name.lower()
                if lower_name not in self.hop_by_hop_headers:
                    response.headers[name] = value
                elif response.status == 206 and lower_name == "content-length":
                    content_length = int(value)
                    start = 0 if range_start is None else range_start
                    stop = content_length if range_stop is None else range_stop

                    range_bytes = stop - start
                    response.headers[name] = str(range_bytes)

                    # aiohttp adds a 1 to the range.stop compared to http headers (including) to
                    # match python array adressing (exclusive)
                    response.headers["Content-Range"] = "bytes {0}-{1}/{2}".format(
                        start, stop - 1, content_length
                    )
            await response.prepare(request)

        data_size_handled = 0

        async def handle_data(data):
            nonlocal data_size_handled
            if range_start or range_stop:
                start_byte_pos = 0
                end_byte_pos = len(data)
                if range_start:
                    start_byte_pos = max(0, range_start - data_size_handled)
                if range_stop:
                    end_byte_pos = min(len(data), range_stop - data_size_handled)

                data_for_client = data[start_byte_pos:end_byte_pos]
                await response.write(data_for_client)
                data_size_handled = data_size_handled + len(data)
            else:
                await response.write(data)
            if remote.policy != Remote.STREAMED:
                await original_handle_data(data)

        async def finalize():
            if remote.policy != Remote.STREAMED:
                await original_finalize()

        downloader = remote.get_downloader(
            remote_artifact=remote_artifact, headers_ready_callback=handle_response_headers
        )
        original_handle_data = downloader.handle_data
        downloader.handle_data = handle_data
        original_finalize = downloader.finalize
        downloader.finalize = finalize
        download_result = await downloader.run()

        if remote.policy != Remote.STREAMED:
            await asyncio.shield(
                sync_to_async(self._save_artifact)(download_result, remote_artifact)
            )
        await response.write_eof()

        if response.status == 404:
            raise HTTPNotFound()
        return response



import os
import io
import json
import tarfile
import tempfile
import logging

from django.conf import settings
from django.db.models.query import QuerySet

from pulpcore.app.apps import get_plugin_config
from pulpcore.app.models.progress import ProgressReport
from pulpcore.app.models.repository import Repository
from pulpcore.app.modelresource import (
    ArtifactResource,
    ContentArtifactResource,
    RepositoryResource,
)
from pulpcore.constants import TASK_STATES, EXPORT_BATCH_SIZE

log = logging.getLogger(__name__)


def _write_export(the_tarfile, resource, dest_dir=None):
    """
    Write the JSON export for the specified resource to the specified tarfile.

    The resulting file will be found at <dest_dir>/<resource.__class__.__name__>.json. If dest_dir
    is None, the file will be added at the 'top level' of the_tarfile.

    Export-files are UTF-8 encoded.

    Args:
        the_tarfile (tarfile.Tarfile): tarfile we are writing into
        resource (import_export.resources.ModelResource): ModelResource to be exported
        dest_dir str(directory-path): directory 'inside' the tarfile to write to
    """
    filename = "{}.{}.json".format(resource.__module__, type(resource).__name__)
    if dest_dir:
        dest_filename = os.path.join(dest_dir, filename)
    else:
        dest_filename = filename

    # If the resource is the type of QuerySet, then export the data in batch to save memory.
    # Otherwise, export all data in oneshot. This is because the underlying libraries
    # (json; django-import-export) do not support to stream the output to file, we export
    # the data in batches to memory and concatenate the json lists via string manipulation.
    with tempfile.NamedTemporaryFile(dir=".", mode="w", encoding="utf8") as temp_file:
        if isinstance(resource.queryset, QuerySet):
            temp_file.write("[")
            total = resource.queryset.count()
            for i in range(0, total, EXPORT_BATCH_SIZE):
                current_batch = i + EXPORT_BATCH_SIZE
                dataset = resource.export(resource.queryset[i:current_batch])
                # Strip "[" and "]" as we are writing the dataset in batch
                temp_file.write(dataset.json.lstrip("[").rstrip("]"))
                if current_batch < total:
                    # Write "," if not last loop
                    temp_file.write(", ")
            temp_file.write("]")
        else:
            dataset = resource.export(resource.queryset)
            temp_file.write(dataset.json)

        temp_file.flush()
        info = tarfile.TarInfo(name=dest_filename)
        info.size = os.path.getsize(temp_file.name)
        with open(temp_file.name, "rb") as fd:
            the_tarfile.addfile(info, fd)


def export_versions(export, version_info):
    """
    Write a JSON list of plugins and their versions as 'versions.json' to export.tarfile

    Output format is [{"component": "<pluginname>", "version": "<pluginversion>"},...]

    Args:
        export (django.db.models.PulpExport): export instance that's doing the export
        version_info (set): set of (distribution-label,version) tuples for repos in this export
    """
    # build the version-list from the distributions for each component
    versions = [{"component": label, "version": version} for (label, version) in version_info]

    version_json = json.dumps(versions).encode("utf8")
    info = tarfile.TarInfo(name="versions.json")
    info.size = len(version_json)
    export.tarfile.addfile(info, io.BytesIO(version_json))


def export_artifacts(export, artifacts):
    """
    Export a set of Artifacts, ArtifactResources, and RepositoryResources

    Args:
        export (django.db.models.PulpExport): export instance that's doing the export
        artifacts (django.db.models.Artifacts): list of artifacts in all repos being exported

    Raises:
        ValidationError: When path is not in the ALLOWED_EXPORT_PATHS setting
    """
    data = dict(message="Exporting Artifacts", code="export.artifacts", total=len(artifacts))
    with ProgressReport(**data) as pb:
        for artifact in pb.iter(artifacts):
            dest = artifact.file.name
            if settings.DEFAULT_FILE_STORAGE != "pulpcore.app.models.storage.FileSystem":
                with tempfile.TemporaryDirectory(dir=".") as temp_dir:
                    with tempfile.NamedTemporaryFile(dir=temp_dir) as temp_file:
                        temp_file.write(artifact.file.read())
                        temp_file.flush()
                        artifact.file.close()
                        export.tarfile.add(temp_file.name, dest)
            else:
                export.tarfile.add(artifact.file.path, dest)

    resource = ArtifactResource()
    resource.queryset = artifacts
    _write_export(export.tarfile, resource)

    resource = RepositoryResource()
    resource.queryset = Repository.objects.filter(pk__in=export.exporter.repositories.all())
    _write_export(export.tarfile, resource)


def export_content(export, repository_version):
    """
    Export db-content, and the db-content of the owning repositories

    Args:
        export (django.db.models.PulpExport): export instance that's doing the export
        repository_version (django.db.models.RepositoryVersion): RepositoryVersion being exported
    """

    def _combine_content_mappings(map1, map2):
        """Combine two content mapping dicts into one by combining ids for for each key."""
        result = {}
        for key in map1.keys() | map2.keys():
            result[key] = list(set(map1.get(key, []) + map2.get(key, [])))
        return result

    dest_dir = os.path.join(
        "repository-{}_{}".format(
            str(repository_version.repository.name), repository_version.number
        )
    )

    # content mapping is used by repo versions with subrepos (eg distribution tree repos)
    content_mapping = {}

    # find and export any ModelResource found in pulp_<repo-type>.app.modelresource
    plugin_name = repository_version.repository.pulp_type.split(".")[0]
    cfg = get_plugin_config(plugin_name)
    if cfg.exportable_classes:
        for cls in cfg.exportable_classes:
            resource = cls(repository_version)
            _write_export(export.tarfile, resource, dest_dir)

            if hasattr(resource, "content_mapping") and resource.content_mapping:
                content_mapping = _combine_content_mappings(
                    content_mapping, resource.content_mapping
                )

    # Export the connection between content and artifacts
    resource = ContentArtifactResource(repository_version, content_mapping)
    _write_export(export.tarfile, resource, dest_dir)

    msg = (
        f"Exporting content for {plugin_name} "
        f"repository-version {repository_version.repository.name}/{repository_version.number}"
    )
    content_count = repository_version.content.count()
    data = dict(
        message=msg,
        code="export.repo.version.content",
        total=content_count,
        done=content_count,
        state=TASK_STATES.COMPLETED,
    )
    pb = ProgressReport(**data)
    pb.save()

    if content_mapping:
        # write the content mapping to tarfile
        cm_json = json.dumps(content_mapping).encode("utf8")
        info = tarfile.TarInfo(name=f"{dest_dir}/content_mapping.json")
        info.size = len(cm_json)
        export.tarfile.addfile(info, io.BytesIO(cm_json))



# Generated by Django 3.2.9 on 2021-11-25 11:20

import django.contrib.auth.models
from django.db import migrations, models
import django.db.models.deletion
import django_lifecycle.mixins
import pulpcore.app.models.access_policy


class Migration(migrations.Migration):

    dependencies = [
        ('auth', '0012_alter_user_first_name_max_length'),
        ('core', '0079_rename_permissions_assignment_accesspolicy_creation_hooks'),
    ]

    operations = [
        migrations.CreateModel(
            name='Group',
            fields=[
            ],
            options={
                'proxy': True,
                'indexes': [],
                'constraints': [],
            },
            bases=(django_lifecycle.mixins.LifecycleModelMixin, 'auth.group', pulpcore.app.models.access_policy.AutoAddObjPermsMixin),
            managers=[
                ('objects', django.contrib.auth.models.GroupManager()),
            ],
        ),
        migrations.AlterField(
            model_name='grouprole',
            name='group',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='object_roles', to='core.group'),
        ),
    ]



"""Tests related to the workers."""
import unittest
from datetime import datetime, timedelta
from random import choice

from pulp_smash import api, config
from pulp_smash.pulp3.constants import WORKER_PATH
from requests.exceptions import HTTPError

from pulpcore.tests.functional.utils import set_up_module as setUpModule  # noqa:F401
from pulpcore.tests.functional.utils import skip_if

_DYNAMIC_WORKER_ATTRS = ("last_heartbeat",)
"""Worker attributes that are dynamically set by Pulp, not set by a user."""


class WorkersTestCase(unittest.TestCase):
    """Test actions over workers.

    This test targets the following issues:

    * `Pulp #3143 <https://pulp.plan.io/issues/3143>`_
    * `Pulp Smash #945 <https://github.com/pulp/pulp-smash/issues/945>`_
    """

    @classmethod
    def setUpClass(cls):
        """Create an API Client."""
        cls.cfg = config.get_config()
        cls.client = api.Client(cls.cfg, api.json_handler)
        cls.worker = {}

    def test_01_read_all_workers(self):
        """Read all workers.

        Pick a random worker to be used for the next assertions.
        """
        workers = self.client.get(WORKER_PATH)["results"]
        for worker in workers:
            for key, val in worker.items():
                with self.subTest(key=key):
                    self.assertIsNotNone(val)
        self.worker.update(choice(workers))

    @skip_if(bool, "worker", False)
    def test_02_read_worker(self):
        """Read a worker by its pulp_href."""
        worker = self.client.get(self.worker["pulp_href"])
        for key, val in self.worker.items():
            if key in _DYNAMIC_WORKER_ATTRS:
                continue
            with self.subTest(key=key):
                self.assertEqual(worker[key], val)

    @skip_if(bool, "worker", False)
    def test_02_read_workers(self):
        """Read a worker by its name."""
        page = self.client.get(WORKER_PATH, params={"name": self.worker["name"]})
        self.assertEqual(len(page["results"]), 1)
        for key, val in self.worker.items():
            if key in _DYNAMIC_WORKER_ATTRS:
                continue
            with self.subTest(key=key):
                self.assertEqual(page["results"][0][key], val)

    @skip_if(bool, "worker", False)
    def test_03_positive_filters(self):
        """Read a worker using a set of query parameters."""
        page = self.client.get(
            WORKER_PATH,
            params={
                "last_heartbeat__gte": self.worker["last_heartbeat"],
                "name": self.worker["name"],
            },
        )
        self.assertEqual(
            len(page["results"]), 1, "Expected: {}. Got: {}.".format([self.worker], page["results"])
        )
        for key, val in self.worker.items():
            if key in _DYNAMIC_WORKER_ATTRS:
                continue
            with self.subTest(key=key):
                self.assertEqual(page["results"][0][key], val)

    @skip_if(bool, "worker", False)
    def test_04_negative_filters(self):
        """Read a worker with a query that does not match any worker."""
        page = self.client.get(
            WORKER_PATH,
            params={
                "last_heartbeat__gte": str(datetime.now() + timedelta(days=1)),
                "name": self.worker["name"],
            },
        )
        self.assertEqual(len(page["results"]), 0)

    @skip_if(bool, "worker", False)
    def test_05_http_method(self):
        """Use an HTTP method different than GET.

        Assert an error is raised.
        """
        with self.assertRaises(HTTPError):
            self.client.delete(self.worker["pulp_href"])



from gettext import gettext as _
from django_currentuser.middleware import get_current_authenticated_user
from pulpcore.app.models import (
    ProgressReport,
    Task,
)
from pulpcore.app.role_util import get_objects_for_user
from pulpcore.constants import TASK_STATES

# Delete 1K at a time - better to use less memory, and take a little longer, with a utility
# function like this.
DELETE_LIMIT = 1000
# Key that delete() returns for Tasks
TASK_KEY = "core.Task"


def _details_reporting(current_reports, current_details, totals_pb):
    """
    Create and update progress-reports for each detail-key returned from a delete() call.

    We don't know how many entities will be deleted via cascade-delete until we're all done.

    The function has one special case: we know how many Tasks we're expecting to delete right
    from the beginning. Therefore, we "assume" that the key `core.Task` has been pre-seeded
    with a ProgressReport whose total is correct, in advance, and therefore don't update
    total for that key.

    Args:
        current_reports (dict): key:ProgressReport to record into
    Returns:
        updated current_reports
    """
    entity_count = 0
    for key, curr_detail in current_details.items():
        entity_count += current_details[key]
        if key in current_reports:
            current_reports[key].increase_by(curr_detail)
        else:
            pb = ProgressReport(
                message=_("Purged task-objects of type {}".format(key)),
                code="purge.tasks.key.{}".format(key),
                total=None,
                done=curr_detail,
            )
            pb.save()
            current_reports[key] = pb
    # Update/save totals once
    totals_pb.increase_by(entity_count)
    return current_reports


def purge(finished_before, states):
    """
    This task purges from the database records of tasks which finished prior to the specified time.

    It will remove only tasks that are 'owned' by the current-user (admin-users own All The Things,
    so admins can delete all tasks).

    It will not remove tasks that are incomplete (ie, in states running|waiting|cancelling).

    It reports (using ProgressReport) the total entities deleted, as well as individual counts
    for each class of entity. This shows the results of cascading-deletes that are triggered
    by deleting a Task.

    Args:
        finished_before (DateTime): Earliest finished-time to **NOT** purge.
        states (List[str]): List of task-states we want to purge.

    """
    current_user = get_current_authenticated_user()
    # Tasks, prior to the specified date, in the specified state, owned by the current-user
    tasks_qs = Task.objects.filter(finished_at__lt=finished_before, state__in=states)
    candidate_qs = get_objects_for_user(current_user, "core.delete_task", qs=tasks_qs)
    delete_qs = get_objects_for_user(current_user, "core.delete_task", qs=tasks_qs[:DELETE_LIMIT])
    # Progress bar reporting total-units
    totals_pb = ProgressReport(
        message=_("Purged task-related-objects total"),
        total=None,
        code="purge.tasks.total",
        done=0,
    )
    totals_pb.save()
    # Dictionary to hold progress-reports by delete-details-key
    details_reports = {}

    # Figure out how many Tasks owned by the current user we're about to delete
    expected_total = candidate_qs.count()
    # Build and save a progress-report for that detail
    pb = ProgressReport(
        message=_("Purged task-objects of type {}".format(TASK_KEY)),
        total=expected_total,
        code="purge.tasks.key.{}".format(TASK_KEY),
        done=0,
    )
    pb.save()
    details_reports[TASK_KEY] = pb

    # Our delete-query is going to deal with "the first DELETE_LIMIT tasks that match our
    # criteria", looping until we've deleted everything that fits our parameters
    units_deleted, details = delete_qs.delete()
    # Until our query returns "No tasks deleted", add results into totals and Do It Again
    while units_deleted > 0:
        _details_reporting(details_reports, details, totals_pb)
        units_deleted, details = delete_qs.delete()

    # Complete the progress-reports for the specific entities deleted
    for key, pb in details_reports.items():
        pb.total = pb.done
        pb.state = TASK_STATES.COMPLETED
        pb.save()

    # Complete the totals-ProgressReport
    totals_pb.total = totals_pb.done
    totals_pb.state = TASK_STATES.COMPLETED
    totals_pb.save()



# Generated by Django 2.2.12 on 2020-05-04 04:41

from django.db import migrations, models


def set_existing_task_groups_finished(apps, schema_editor):
    TaskGroup = apps.get_model('core', 'TaskGroup')
    TaskGroup.objects.all().update(all_tasks_dispatched=True)


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0029_export_delete'),
    ]

    operations = [
        migrations.AddField(
            model_name='taskgroup',
            name='all_tasks_dispatched',
            field=models.BooleanField(default=False),
        ),
        migrations.RunPython(set_existing_task_groups_finished, reverse_code=migrations.RunPython.noop, elidable=True),
    ]



# Generated by Django 2.2.20 on 2021-04-29 14:33

from django.db import migrations
import pulpcore.app.models.fields

fields = ("username", "password", "proxy_username", "proxy_password", "client_key")
new_fields = ("_encrypted_username", "_encrypted_password", "_encrypted_proxy_username", "_encrypted_proxy_password", "_encrypted_client_key")


def encrypt_remote_fields(apps, schema_editor):
    Remote = apps.get_model("core", "Remote")

    remotes_needing_update = []
    for remote in Remote.objects.all().iterator():
        if not any([getattr(remote, field) for field in fields]):
            continue

        remote._encrypted_username = remote.username
        remote._encrypted_password = remote.password
        remote._encrypted_proxy_username = remote.proxy_username
        remote._encrypted_proxy_password = remote.proxy_password
        remote._encrypted_client_key = remote.client_key
        remotes_needing_update.append(remote)

        if len(remotes_needing_update) > 100:
            Remote.objects.bulk_update(remotes_needing_update, new_fields)
            remotes_needing_update.clear()

    Remote.objects.bulk_update(remotes_needing_update, new_fields)


def unencrypt_remote_fields(apps, schema_editor):
    Remote = apps.get_model("core", "Remote")

    remotes_needing_update = []
    for remote in Remote.objects.all().iterator():
        if not any([getattr(remote, field) for field in new_fields]):
            continue
        remote.username = remote._encrypted_username
        remote.password = remote._encrypted_password
        remote.proxy_username = remote._encrypted_proxy_username
        remote.proxy_password = remote._encrypted_proxy_password
        remote.client_key = remote._encrypted_client_key
        remotes_needing_update.append(remote)

        if len(remotes_needing_update) > 100:
            Remote.objects.bulk_update(remotes_needing_update, fields)
            remotes_needing_update.clear()

    Remote.objects.bulk_update(remotes_needing_update, fields)


class Migration(migrations.Migration):

    dependencies = [
        ("core", "0072_add_method_to_filesystem_exporter"),
    ]

    operations = [
        # Add new fields to temporarily hold the encrypted values
        migrations.AddField(
            model_name="remote",
            name="_encrypted_client_key",
            field=pulpcore.app.models.fields.EncryptedTextField(null=True),
        ),
        migrations.AddField(
            model_name="remote",
            name="_encrypted_password",
            field=pulpcore.app.models.fields.EncryptedTextField(null=True),
        ),
        migrations.AddField(
            model_name="remote",
            name="_encrypted_proxy_password",
            field=pulpcore.app.models.fields.EncryptedTextField(null=True),
        ),
        migrations.AddField(
            model_name="remote",
            name="_encrypted_proxy_username",
            field=pulpcore.app.models.fields.EncryptedTextField(null=True),
        ),
        migrations.AddField(
            model_name="remote",
            name="_encrypted_username",
            field=pulpcore.app.models.fields.EncryptedTextField(null=True),
        ),
        # Populate the new fields with encrypted values computed from the unencrypted fields
        migrations.RunPython(
            code=encrypt_remote_fields,
            reverse_code=unencrypt_remote_fields,
            elidable=True,
        ),
        # Remove the unencrypted columns
        migrations.RemoveField(
            model_name="remote",
            name="client_key",
        ),
        migrations.RemoveField(
            model_name="remote",
            name="password",
        ),
        migrations.RemoveField(
            model_name="remote",
            name="proxy_password",
        ),
        migrations.RemoveField(
            model_name="remote",
            name="proxy_username",
        ),
        migrations.RemoveField(
            model_name="remote",
            name="username",
        ),
        # Replace the formerly-unencrypted columns with the new encrypted ones
        migrations.RenameField(
            model_name="remote",
            old_name="_encrypted_client_key",
            new_name="client_key",
        ),
        migrations.RenameField(
            model_name="remote",
            old_name="_encrypted_password",
            new_name="password",
        ),
        migrations.RenameField(
            model_name="remote",
            old_name="_encrypted_proxy_password",
            new_name="proxy_password",
        ),
        migrations.RenameField(
            model_name="remote",
            old_name="_encrypted_proxy_username",
            new_name="proxy_username",
        ),
        migrations.RenameField(
            model_name="remote",
            old_name="_encrypted_username",
            new_name="username",
        ),
    ]



from django_filters.rest_framework import filters
from rest_framework import mixins
from rest_framework.decorators import action
from rest_framework.response import Response
from drf_spectacular.utils import extend_schema

from pulpcore.app.apps import pulp_plugin_configs
from pulpcore.app.models import AccessPolicy
from pulpcore.app.serializers import AccessPolicySerializer
from pulpcore.app.viewsets import BaseFilterSet, NamedModelViewSet
from pulpcore.app.viewsets.base import NAME_FILTER_OPTIONS
from pulpcore.app.util import get_view_urlpattern


class AccessPolicyFilter(BaseFilterSet):
    """
    FilterSet for AccessPolicy.
    """

    customized = filters.BooleanFilter()

    class Meta:
        model = AccessPolicy
        fields = {"viewset_name": NAME_FILTER_OPTIONS, "customized": ["exact"]}


class AccessPolicyViewSet(
    NamedModelViewSet, mixins.UpdateModelMixin, mixins.RetrieveModelMixin, mixins.ListModelMixin
):
    """
    ViewSet for AccessPolicy.

    NOTE: This API endpoint is in "tech preview" and subject to change

    """

    queryset = AccessPolicy.objects.all()
    serializer_class = AccessPolicySerializer
    endpoint_name = "access_policies"
    ordering = "viewset_name"
    filterset_class = AccessPolicyFilter

    @extend_schema(request=None)
    @action(detail=True, methods=["post"])
    def reset(self, request, pk=None):
        """
        Reset the access policy to its uncustomized default value.
        """

        access_policy = self.get_object()
        for plugin_config in pulp_plugin_configs():
            for viewset_batch in plugin_config.named_viewsets.values():
                for viewset in viewset_batch:
                    if get_view_urlpattern(viewset) == access_policy.viewset_name:
                        default_access_policy = viewset.DEFAULT_ACCESS_POLICY
                        access_policy.statements = default_access_policy["statements"]
                        access_policy.creation_hooks = default_access_policy.get(
                            "creation_hooks"
                        ) or default_access_policy.get("permissions_assignment")
                        access_policy.customized = False
                        access_policy.save()
                        serializer = AccessPolicySerializer(
                            access_policy, context={"request": request}
                        )
                        return Response(serializer.data)
        raise RuntimeError("Viewset for access policy was not found.")



# Generated by Django 2.2.24 on 2021-06-13 03:14

import django.core.validators
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0065_merge_20210615_1211'),
    ]

    operations = [
        migrations.AlterField(
            model_name='remote',
            name='download_concurrency',
            field=models.PositiveIntegerField(null=True, validators=[django.core.validators.MinValueValidator(1, 'Download concurrency must be at least 1')]),
        ),
        migrations.AddField(
            model_name='remote',
            name='max_retries',
            field=models.PositiveIntegerField(null=True),
        ),
    ]



from pulp_smash.pulp3.bindings import delete_orphans, monitor_task, PulpTestCase
from pulp_smash.pulp3.utils import gen_repo

from pulpcore.tests.functional.api.using_plugin.utils import (
    gen_file_client,
    gen_file_remote,
)
from pulpcore.client.pulp_file import (
    ContentFilesApi,
    RepositorySyncURL,
    RepositoriesFileApi,
    RemotesFileApi,
)
from pulpcore.tests.functional.api.using_plugin.utils import (  # noqa:F401
    set_up_module as setUpModule,
)


class MultiplePolicySyncTestCase(PulpTestCase):
    """
    This test ensures that content artifacts are properly updated when syncing multiple
    times with different policies, specifically from 'on_demand' to 'immediate'

    This test targets the following issue:
    * `Pulp #9101 <https://pulp.plan.io/issues/9101>`_
    """

    @classmethod
    def setUpClass(cls):
        """Clean out Pulp before testing."""
        delete_orphans()
        client = gen_file_client()
        cls.cont_api = ContentFilesApi(client)
        cls.repo_api = RepositoriesFileApi(client)
        cls.remote_api = RemotesFileApi(client)

    def tearDown(self):
        """Clean up Pulp after testing."""
        self.doCleanups()
        delete_orphans()

    def test_ondemand_to_immediate_sync(self):
        """Checks that content artifacts are updated following on-demand -> immediate sync."""
        # Ensure that no content is present
        content_response = self.cont_api.list(limit=1)
        if content_response.count > 0:
            self.skipTest("Please remove all file content before running this test")

        # Create and sync repo w/ on_demand policy
        repo = self.repo_api.create(gen_repo())
        remote = self.remote_api.create(gen_file_remote(policy="on_demand"))
        body = RepositorySyncURL(remote=remote.pulp_href)
        monitor_task(self.repo_api.sync(repo.pulp_href, body).task)
        self.addCleanup(self.repo_api.delete, repo.pulp_href)
        self.addCleanup(self.remote_api.delete, remote.pulp_href)

        # Check content is present, but no artifacts are there
        content_response = self.cont_api.list()
        self.assertEqual(content_response.count, 3)
        for content in content_response.results:
            self.assertEqual(content.artifact, None)

        # Sync again w/ immediate policy
        remote = self.remote_api.create(gen_file_remote())
        body = RepositorySyncURL(remote=remote.pulp_href)
        monitor_task(self.repo_api.sync(repo.pulp_href, body).task)
        self.addCleanup(self.remote_api.delete, remote.pulp_href)

        # Check content is still present, but artifacts are now there
        content_response = self.cont_api.list()
        self.assertEqual(content_response.count, 3)
        for content in content_response.results:
            self.assertNotEqual(content.artifact, None)



# Generated by Django 2.2.6 on 2019-10-07 16:45

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0008_published_metadata_as_content'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='task',
            name='non_fatal_errors',
        ),
    ]



from django_filters.rest_framework import filters

from drf_spectacular.utils import extend_schema
from rest_framework import mixins

from pulpcore.app.models import (
    Export,
    Exporter,
    FilesystemExport,
    FilesystemExporter,
    Publication,
    PulpExport,
    PulpExporter,
    RepositoryVersion,
)

from pulpcore.app.serializers import (
    AsyncOperationResponseSerializer,
    ExportSerializer,
    ExporterSerializer,
    FilesystemExporterSerializer,
    FilesystemExportSerializer,
    PulpExporterSerializer,
    PulpExportSerializer,
)

from pulpcore.app.tasks.export import fs_publication_export, fs_repo_version_export, pulp_export

from pulpcore.app.viewsets import (
    AsyncRemoveMixin,
    AsyncUpdateMixin,
    BaseFilterSet,
    NamedModelViewSet,
)
from pulpcore.app.viewsets.base import NAME_FILTER_OPTIONS
from pulpcore.plugin.tasking import dispatch
from pulpcore.app.response import OperationPostponedResponse


class ExporterFilter(BaseFilterSet):
    """
    Plugin file system exporter filter should:
     - inherit from this class
     - add any specific filters if needed
     - define a `Meta` class which should:
       - specify a plugin remote model for which filter is defined
       - extend `fields` with specific ones
    """

    name = filters.CharFilter()

    class Meta:
        model = Exporter
        fields = {
            "name": NAME_FILTER_OPTIONS,
        }


class ExporterViewSet(
    NamedModelViewSet,
    mixins.CreateModelMixin,
    AsyncUpdateMixin,
    mixins.RetrieveModelMixin,
    mixins.ListModelMixin,
    AsyncRemoveMixin,
):
    """
    ViewSet for viewing exporters.
    """

    queryset = Exporter.objects.all()
    serializer_class = ExporterSerializer
    endpoint_name = "exporters"
    router_lookup = "exporter"
    filterset_class = ExporterFilter


class PulpExporterViewSet(ExporterViewSet):
    """
    ViewSet for viewing PulpExporters.
    """

    endpoint_name = "pulp"
    serializer_class = PulpExporterSerializer
    queryset = PulpExporter.objects.all()


class FilesystemExporterViewSet(ExporterViewSet):
    """
    Endpoint for managing FilesystemExporters. FilesystemExporters are provided as a tech preview.
    """

    endpoint_name = "filesystem"
    serializer_class = FilesystemExporterSerializer
    queryset = FilesystemExporter.objects.all()


class ExportViewSet(
    NamedModelViewSet,
    mixins.CreateModelMixin,
    mixins.RetrieveModelMixin,
    mixins.ListModelMixin,
    mixins.DestroyModelMixin,
):
    """
    ViewSet for viewing exports from an Exporter.
    """

    endpoint_name = "exports"
    nest_prefix = "exporters"
    router_lookup = "export"
    lookup_field = "pk"
    parent_lookup_kwargs = {"exporter_pk": "exporter__pk"}
    serializer_class = ExportSerializer
    queryset = Export.objects.all()
    parent_viewset = ExporterViewSet


class PulpExportViewSet(ExportViewSet):
    """
    ViewSet for viewing exports from a PulpExporter.
    """

    parent_viewset = PulpExporterViewSet
    serializer_class = PulpExportSerializer
    queryset = PulpExport.objects.all()

    @extend_schema(
        request=PulpExportSerializer,
        description="Trigger an asynchronous task to export a set of repositories",
        responses={202: AsyncOperationResponseSerializer},
    )
    def create(self, request, exporter_pk):
        """
        Generates a Task to export the set of repositories assigned to a specific PulpExporter.
        """
        # Validate Exporter
        exporter = PulpExporter.objects.get(pk=exporter_pk).cast()
        ExporterSerializer.validate_path(exporter.path, check_is_dir=True)

        # Validate Export
        serializer = PulpExportSerializer(data=request.data, context={"exporter": exporter})
        serializer.is_valid(raise_exception=True)

        # Invoke the export
        task = dispatch(
            pulp_export,
            exclusive_resources=[exporter],
            kwargs={"exporter_pk": str(exporter.pk), "params": request.data},
        )

        return OperationPostponedResponse(task, request)


class FilesystemExportViewSet(ExportViewSet):
    """
    Endpoint for managing FilesystemExports. This endpoint is provided as a tech preview.
    """

    parent_viewset = FilesystemExporterViewSet
    serializer_class = FilesystemExportSerializer
    queryset = FilesystemExport.objects.all()

    @extend_schema(
        request=FilesystemExportSerializer,
        description="Trigger an asynchronous task to export files to the filesystem",
        responses={202: AsyncOperationResponseSerializer},
    )
    def create(self, request, exporter_pk):
        """
        Generates a Task to export files to the filesystem.
        """
        # Validate Exporter
        exporter = FilesystemExporter.objects.get(pk=exporter_pk).cast()
        ExporterSerializer.validate_path(exporter.path, check_is_dir=True)

        # Validate Export
        serializer = FilesystemExportSerializer(data=request.data, context={"exporter": exporter})
        serializer.is_valid(raise_exception=True)

        if request.data.get("publication"):
            publication = self.get_resource(request.data["publication"], Publication)

            task = dispatch(
                fs_publication_export,
                exclusive_resources=[exporter],
                kwargs={"exporter_pk": exporter.pk, "publication_pk": publication.pk},
            )
        else:
            repo_version = self.get_resource(request.data["repository_version"], RepositoryVersion)

            task = dispatch(
                fs_repo_version_export,
                exclusive_resources=[exporter],
                kwargs={"exporter_pk": str(exporter.pk), "repo_version_pk": repo_version.pk},
            )

        return OperationPostponedResponse(task, request)



# Load order: base, fields, all others.
# - fields can import directly from base if needed
# - all can import directly from base and fields if needed

from .base import (  # noqa
    AsyncOperationResponseSerializer,
    DetailIdentityField,
    DetailRelatedField,
    IdentityField,
    ModelSerializer,
    NestedIdentityField,
    NestedRelatedField,
    RelatedField,
    RelatedResourceField,
    TaskGroupOperationResponseSerializer,
    ValidateFieldsMixin,
    validate_unknown_fields,
)
from .fields import (  # noqa
    BaseURLField,
    ExportsIdentityFromExporterField,
    ExportRelatedField,
    ExportIdentityField,
    ImportsIdentityFromImporterField,
    ImportRelatedField,
    ImportIdentityField,
    LabelsField,
    LatestVersionField,
    SingleContentArtifactField,
    RepositoryVersionsIdentityFromRepositoryField,
    RepositoryVersionRelatedField,
    RepositoryVersionIdentityField,
    relative_path_validator,
    TaskGroupStatusCountField,
)
from .access_policy import AccessPolicySerializer  # noqa
from .acs import (  # noqa
    AlternateContentSourcePathSerializer,
    AlternateContentSourceSerializer,
)
from .content import (  # noqa
    ArtifactSerializer,
    ContentChecksumSerializer,
    MultipleArtifactContentSerializer,
    NoArtifactContentSerializer,
    SigningServiceSerializer,
    SingleArtifactContentSerializer,
)
from .exporter import (  # noqa
    ExporterSerializer,
    ExportSerializer,
    FilesystemExportSerializer,
    FilesystemExporterSerializer,
    PulpExporterSerializer,
    PulpExportSerializer,
)
from .importer import (  # noqa
    EvaluationSerializer,
    ImporterSerializer,
    ImportSerializer,
    PulpImportCheckResponseSerializer,
    PulpImportCheckSerializer,
    PulpImporterSerializer,
    PulpImportSerializer,
)
from .orphans import OrphansCleanupSerializer  # noqa
from .progress import GroupProgressReportSerializer, ProgressReportSerializer  # noqa
from .publication import (  # noqa
    ContentGuardSerializer,
    DistributionSerializer,
    PublicationSerializer,
    RBACContentGuardSerializer,
    RBACContentGuardPermissionSerializer,
)
from .purge import PurgeSerializer  # noqa
from .repository import (  # noqa
    RemoteSerializer,
    RepositorySerializer,
    RepositorySyncURLSerializer,
    RepositoryAddRemoveContentSerializer,
    RepositoryVersionSerializer,
)
from .repair import RepairSerializer  # noqa
from .reclaim import ReclaimSpaceSerializer  # noqa
from .task import (  # noqa
    MinimalTaskSerializer,
    TaskCancelSerializer,
    TaskSerializer,
    TaskGroupSerializer,
    WorkerSerializer,
)
from .upload import (  # noqa
    UploadChunkSerializer,
    UploadCommitSerializer,
    UploadSerializer,
    UploadDetailSerializer,
)
from .user import (  # noqa
    GroupRoleSerializer,
    GroupSerializer,
    GroupUserSerializer,
    NestedRoleSerializer,
    PermissionSerializer,
    RoleSerializer,
    UserRoleSerializer,
    UserSerializer,
)



"""
Tests PulpExporter and PulpExport functionality

NOTE: assumes ALLOWED_EXPORT_PATHS setting contains "/tmp" - all tests will fail if this is not
the case.
"""
import unittest
from pulp_smash import api, cli, config, utils
from pulp_smash.utils import uuid4
from pulp_smash.pulp3.bindings import delete_orphans, monitor_task
from pulp_smash.pulp3.utils import (
    gen_repo,
)

from pulpcore.tests.functional.api.using_plugin.utils import (
    create_repo_and_versions,
    delete_exporter,
    gen_file_client,
    gen_file_remote,
)

from pulpcore.client.pulpcore import (
    ApiClient as CoreApiClient,
    ExportersPulpApi,
    ExportersPulpExportsApi,
)

from pulpcore.client.pulpcore.exceptions import ApiException

from pulpcore.client.pulp_file import (
    ContentFilesApi,
    RepositoriesFileApi,
    RepositoriesFileVersionsApi,
    RepositorySyncURL,
    RemotesFileApi,
)
from pulpcore.constants import TASK_STATES

NUM_REPOS = 3
MAX_EXPORTS = 3
NUM_EXPORTERS = 4


class BaseExporterCase(unittest.TestCase):
    """
    Base functionality for Exporter and Export test classes

    The export process isn't possible without repositories having been sync'd - arranging for
    that to happen once per-class (instead of once-per-test) is the primary purpose of this parent
    class.
    """

    @classmethod
    def _setup_repositories(cls):
        """Create and sync a number of repositories to be exported."""
        # create and remember a set of repo
        repos = []
        remotes = []
        for r in range(NUM_REPOS):
            a_repo = cls.repo_api.create(gen_repo())
            # give it a remote and sync it
            body = gen_file_remote()
            remote = cls.remote_api.create(body)
            repository_sync_data = RepositorySyncURL(remote=remote.pulp_href)
            sync_response = cls.repo_api.sync(a_repo.pulp_href, repository_sync_data)
            monitor_task(sync_response.task)
            # remember it
            a_repo = cls.repo_api.read(file_file_repository_href=a_repo.pulp_href)
            repos.append(a_repo)
            remotes.append(remote)
        return repos, remotes

    @classmethod
    def setUpClass(cls):
        """Create class-wide variables."""

        cls.cfg = config.get_config()
        cls.cli_client = cli.Client(cls.cfg)
        allowed_exports = utils.get_pulp_setting(cls.cli_client, "ALLOWED_EXPORT_PATHS")
        if not allowed_exports or "/tmp" not in allowed_exports:
            raise unittest.SkipTest(
                "Cannot run export-tests unless /tmp is in ALLOWED_EXPORT_PATHS ({}).".format(
                    allowed_exports
                ),
            )

        cls.client = api.Client(cls.cfg, api.json_handler)
        cls.core_client = CoreApiClient(configuration=cls.cfg.get_bindings_config())
        cls.file_client = gen_file_client()

        cls.content_api = ContentFilesApi(cls.file_client)
        cls.repo_api = RepositoriesFileApi(cls.file_client)
        cls.versions_api = RepositoriesFileVersionsApi(cls.file_client)
        cls.remote_api = RemotesFileApi(cls.file_client)
        cls.exporter_api = ExportersPulpApi(cls.core_client)
        cls.exports_api = ExportersPulpExportsApi(cls.core_client)

        (cls.repos, cls.remotes) = cls._setup_repositories()

    @classmethod
    def tearDownClass(cls):
        """Clean up after ourselves."""
        for remote in cls.remotes:
            cls.remote_api.delete(remote.pulp_href)
        for repo in cls.repos:
            cls.repo_api.delete(repo.pulp_href)
        delete_orphans()

    def _create_exporter(self, cleanup=True, use_repos=None):
        """
        Utility routine to create an exporter for the available repositories.

        If all_repos, export everything in self.repos; otherwise only export first repo
        """

        body = {
            "name": uuid4(),
            "path": "/tmp/{}/".format(uuid4()),
            "repositories": [r.pulp_href for r in self.repos],
        }
        if use_repos:
            body["repositories"] = [r.pulp_href for r in use_repos]

        exporter = self.exporter_api.create(body)
        if cleanup:
            self.addCleanup(delete_exporter, exporter)
        return exporter, body


class PulpExporterTestCase(BaseExporterCase):
    """Test PulpExporter CURDL methods."""

    def test_create(self):
        """Create a PulpExporter."""
        (exporter, body) = self._create_exporter()
        self.assertIsNone(exporter.last_export)
        self.assertEqual(body["name"], exporter.name)
        self.assertEqual(body["path"], exporter.path)
        self.assertEqual(len(self.repos), len(exporter.repositories))

    def test_read(self):
        """Read a created PulpExporter."""
        (exporter_created, body) = self._create_exporter()
        exporter_read = self.exporter_api.read(exporter_created.pulp_href)
        self.assertEqual(exporter_created.name, exporter_read.name)
        self.assertEqual(exporter_created.path, exporter_read.path)
        self.assertEqual(len(exporter_created.repositories), len(exporter_read.repositories))

    def test_partial_update(self):
        """Update a PulpExporter's path."""
        (exporter_created, body) = self._create_exporter()
        body = {"path": "/tmp/{}".format(uuid4())}
        result = self.exporter_api.partial_update(exporter_created.pulp_href, body)
        monitor_task(result.task)
        exporter_read = self.exporter_api.read(exporter_created.pulp_href)
        self.assertNotEqual(exporter_created.path, exporter_read.path)
        self.assertEqual(body["path"], exporter_read.path)

    def test_list(self):
        """Show a set of created PulpExporters."""
        starting_exporters = self.exporter_api.list().results
        for x in range(NUM_EXPORTERS):
            self._create_exporter()
        ending_exporters = self.exporter_api.list().results
        self.assertEqual(NUM_EXPORTERS, len(ending_exporters) - len(starting_exporters))

    def test_delete(self):
        """Delete a pulpExporter."""
        (exporter_created, body) = self._create_exporter(cleanup=False)
        delete_exporter(exporter_created)
        try:
            self.exporter_api.read(exporter_created.pulp_href)
        except ApiException as ae:
            self.assertEqual(404, ae.status)
            return
        self.fail("Found a deleted exporter!")


class PulpExportTestCase(BaseExporterCase):
    """Test PulpExport CRDL methods (Update is not allowed)."""

    def _gen_export(self, exporter, body={}):
        """Create and read back an export for the specified PulpExporter."""
        export_response = self.exports_api.create(exporter.pulp_href, body)
        monitor_task(export_response.task)
        task = self.client.get(export_response.task)
        resources = task["created_resources"]
        self.assertEqual(1, len(resources))
        reports = task["progress_reports"]
        found_artifacts = False
        found_content = False
        for r in reports:
            self.assertEqual(TASK_STATES.COMPLETED, r["state"])
            found_artifacts |= r["code"] == "export.artifacts"
            found_content |= r["code"] == "export.repo.version.content"
        self.assertTrue(found_artifacts, "No artifacts exported!")
        self.assertTrue(found_content, "No content exported!")
        export_href = resources[0]
        export = self.exports_api.read(export_href)
        self.assertIsNotNone(export)
        return export

    def test_export(self):
        """Issue and evaluate a PulpExport (tests both Create and Read)."""
        (exporter, body) = self._create_exporter(cleanup=False)
        try:
            export = self._gen_export(exporter)
            self.assertIsNotNone(export)
            self.assertEqual(len(exporter.repositories), len(export.exported_resources))
            self.assertIsNotNone(export.output_file_info)
            self.assertIsNotNone(export.toc_info)
            for an_export_filename in export.output_file_info.keys():
                self.assertFalse("//" in an_export_filename)

        finally:
            delete_exporter(exporter)

    def test_list(self):
        """Find all the PulpExports for a PulpExporter."""
        (exporter, body) = self._create_exporter(cleanup=False)
        try:
            export = None
            for i in range(MAX_EXPORTS):
                export = self._gen_export(exporter)
            exporter = self.exporter_api.read(exporter.pulp_href)
            self.assertEqual(exporter.last_export, export.pulp_href)
            exports = self.exports_api.list(exporter.pulp_href).results
            self.assertEqual(MAX_EXPORTS, len(exports))
        finally:
            delete_exporter(exporter)

    def _delete_export(self, export):
        """
        Delete a PulpExport and test that it is gone.

        :param export: PulpExport to be deleted
        :return: true if specified export is gone, false if we can still find it
        """
        self.exports_api.delete(export.pulp_href)
        try:
            self.exports_api.read(export.pulp_href)
        except ApiException as ae:
            self.assertEqual(404, ae.status)
            return True
        return False

    def _create_repo_and_versions(self):
        a_repo, versions = create_repo_and_versions(
            self.repos[0], self.repo_api, self.versions_api, self.content_api
        )
        self.addCleanup(self.client.delete, a_repo.pulp_href)
        self.assertIsNotNone(versions)
        self.assertEqual(4, versions.count)
        return a_repo, versions

    def test_delete(self):
        """
        Test deleting exports for a PulpExporter.

        NOTE: Attempting to delete the current last_export is forbidden.
        """
        (exporter, body) = self._create_exporter(cleanup=False)
        try:
            # Do three exports
            first_export = self._gen_export(exporter)
            self._gen_export(exporter)
            last_export = self._gen_export(exporter)

            # delete one make sure it's gone
            if not self._delete_export(first_export):
                self.fail("Failed to delete an export")

            # make sure the exporter knows it's gone
            exporter = self.exporter_api.read(exporter.pulp_href)
            exports = self.exports_api.list(exporter.pulp_href).results
            self.assertEqual(2, len(exports))

            # Now try to delete the last_export export and succeed
            # as of https://pulp.plan.io/issues/6555
            self._delete_export(last_export)
            # Make sure the exporter is still around...
            exporter = self.exporter_api.read(exporter.pulp_href)
        finally:
            delete_exporter(exporter)

    @unittest.skip("not yet implemented")
    def test_export_output(self):
        """Create an export and evaluate the resulting export-file."""
        self.fail("test_export_file")

    def test_export_by_version_validation(self):
        repositories = self.repos
        latest_versions = [r.latest_version_href for r in repositories]

        # exporter for one repo. specify one version
        (exporter, body) = self._create_exporter(use_repos=[repositories[0]])
        body = {"versions": [latest_versions[0]]}
        self._gen_export(exporter, body)

        # exporter for one repo. specify one *wrong* version
        with self.assertRaises(ApiException) as ae:
            (exporter, body) = self._create_exporter(use_repos=[repositories[0]])
            body = {"versions": [latest_versions[1]]}
            self._gen_export(exporter, body)
        self.assertTrue("must belong to" in ae.exception.body)

        # exporter for two repos, specify one version
        with self.assertRaises(ApiException) as ae:
            (exporter, body) = self._create_exporter(use_repos=[repositories[0], repositories[1]])
            body = {"versions": [latest_versions[0]]}
            self._gen_export(exporter, body)
        self.assertTrue("does not match the number" in ae.exception.body)

        # exporter for two repos, specify one correct and one *wrong* version
        with self.assertRaises(ApiException) as ae:
            (exporter, body) = self._create_exporter(use_repos=[repositories[0], repositories[1]])
            body = {"versions": [latest_versions[0], latest_versions[2]]}
            self._gen_export(exporter, body)
        self.assertTrue("must belong to" in ae.exception.body)

    def test_export_by_version_results(self):
        repositories = self.repos
        latest_versions = [r.latest_version_href for r in repositories]
        zeroth_versions = []
        for v in latest_versions:
            v_parts = v.split("/")
            v_parts[-2] = "0"
            zeroth_versions.append("/".join(v_parts))

        (exporter, body) = self._create_exporter(use_repos=[repositories[0]], cleanup=False)
        try:
            # export no-version, check that /1/ was exported
            export = self._gen_export(exporter)
            self.assertTrue(export.exported_resources[0].endswith("/1/"))
            # exporter-by-version, check that /0/ was exported
            body = {"versions": [zeroth_versions[0]]}
            export = self._gen_export(exporter, body)
            self.assertTrue(export.exported_resources[0].endswith("/0/"))
        finally:
            delete_exporter(exporter)

    def test_incremental(self):
        # create a repo with 4 repo-versions
        a_repo, versions = self._create_repo_and_versions()
        # create exporter for that repository
        (exporter, body) = self._create_exporter(use_repos=[a_repo], cleanup=False)
        try:
            # negative - ask for an incremental without having a last_export
            with self.assertRaises(ApiException):
                body = {"full": False}
                self._gen_export(exporter, body)

            # export repo-2-version[1]-full versions.results[1]
            body = {"versions": [versions.results[1].pulp_href]}
            self._gen_export(exporter, body)
            # export repo-2-version[2]
            body = {"versions": [versions.results[2].pulp_href], "full": False}
            self._gen_export(exporter, body)
            # export repo-2-latest
            body = {"full": False}
            self._gen_export(exporter, body)
        finally:
            delete_exporter(exporter)

    def test_chunking(self):
        a_repo = self.repo_api.create(gen_repo())
        self.addCleanup(self.client.delete, a_repo.pulp_href)
        (exporter, body) = self._create_exporter(use_repos=[a_repo], cleanup=False)
        try:
            body = {"chunk_size": "250B"}
            export = self._gen_export(exporter, body)
            info = export.output_file_info
            self.assertIsNotNone(info)
            self.assertTrue(len(info) > 1)
        finally:
            delete_exporter(exporter)

    def test_start_end_incrementals(self):
        # create a repo with 4 repo-versions
        a_repo, versions = self._create_repo_and_versions()
        (exporter, body) = self._create_exporter(use_repos=[a_repo], cleanup=False)
        try:
            # export from version-1 to latest last=v3
            body = {"start_versions": [versions.results[1].pulp_href], "full": False}
            self._gen_export(exporter, body)

            # export from version-1 to version-2, last=v2
            body = {
                "start_versions": [versions.results[1].pulp_href],
                "versions": [versions.results[2].pulp_href],
                "full": False,
            }
            self._gen_export(exporter, body)

            # negative attempt, start_versions= is not a version
            with self.assertRaises(ApiException):
                body = {"start_versions": [a_repo.pulp_href], "full": False}
                self._gen_export(exporter, body)

            # negative attempt, start_versions= and Full=True
            with self.assertRaises(ApiException):
                body = {"start_versions": [versions.results[2].pulp_href], "full": True}
                self._gen_export(exporter, body)

            # negative attempt, start_versions= is a version from Some Other Repo
            with self.assertRaises(ApiException):
                second_repo, second_versions = self._create_repo_and_versions()
                body = {"start_versions": [second_versions.results[0].pulp_href], "full": False}
                self._gen_export(exporter, body)
        finally:
            delete_exporter(exporter)



import re

from gettext import gettext as _
from rest_framework import serializers

from pulpcore.app import models
from pulpcore.app.serializers import base, ValidateFieldsMixin


CONTENT_RANGE_PATTERN = r"^bytes (\d+)-(\d+)/(\d+|[*])$"


class UploadChunkSerializer(ValidateFieldsMixin, serializers.Serializer):
    file = serializers.FileField(help_text=_("A chunk of the uploaded file."), write_only=True)

    sha256 = serializers.CharField(
        help_text=_("The SHA-256 checksum of the chunk if available."),
        required=False,
        allow_null=True,
        write_only=True,
    )

    offset = serializers.IntegerField(read_only=True)

    size = serializers.IntegerField(read_only=True)

    def validate(self, data):
        data = super().validate(data)

        content_range = self.context["request"].META.get("HTTP_CONTENT_RANGE", "")
        match = re.compile(CONTENT_RANGE_PATTERN).match(content_range)
        if not match:
            raise serializers.ValidationError(_("Invalid or missing content range header."))
        data["start"] = start = int(match[1])
        end = int(match[2])

        if (end - start + 1) != len(data["file"]):
            raise serializers.ValidationError(_("Chunk size does not match content range."))

        if end > self.context["upload"].size - 1:
            raise serializers.ValidationError(_("End byte is greater than upload size."))

        return data

    class Meta:
        model = models.UploadChunk
        fields = ("file", "sha256", "offset", "size")


class UploadSerializer(base.ModelSerializer):
    """Serializer for chunked uploads."""

    pulp_href = base.IdentityField(view_name="uploads-detail")

    size = serializers.IntegerField(help_text=_("The size of the upload in bytes."))

    completed = serializers.DateTimeField(
        help_text=_("Timestamp when upload is committed."), read_only=True
    )

    class Meta:
        model = models.Upload
        fields = base.ModelSerializer.Meta.fields + ("size", "completed")


class UploadDetailSerializer(UploadSerializer):
    chunks = UploadChunkSerializer(many=True, read_only=True)

    class Meta(UploadSerializer.Meta):
        fields = UploadSerializer.Meta.fields + ("chunks",)


class UploadCommitSerializer(ValidateFieldsMixin, serializers.Serializer):
    sha256 = serializers.CharField(help_text=_("The expected sha256 checksum for the file."))



from gettext import gettext as _

from django_filters import Filter
from django_filters.rest_framework import DjangoFilterBackend, filters
from drf_spectacular.utils import extend_schema
from rest_framework import mixins, serializers
from rest_framework.decorators import action
from rest_framework.filters import OrderingFilter
from rest_framework.viewsets import GenericViewSet

from pulpcore.app import tasks
from pulpcore.app.models import (
    Content,
    Remote,
    Repository,
    RepositoryVersion,
)
from pulpcore.app.response import OperationPostponedResponse
from pulpcore.app.serializers import (
    AsyncOperationResponseSerializer,
    RemoteSerializer,
    RepairSerializer,
    RepositorySerializer,
    RepositoryVersionSerializer,
)
from pulpcore.app.viewsets import (
    AsyncRemoveMixin,
    AsyncUpdateMixin,
    BaseFilterSet,
    NamedModelViewSet,
)
from pulpcore.app.viewsets.base import DATETIME_FILTER_OPTIONS, NAME_FILTER_OPTIONS
from pulpcore.app.viewsets.custom_filters import IsoDateTimeFilter, LabelSelectFilter
from pulpcore.tasking.tasks import dispatch


class RepositoryFilter(BaseFilterSet):
    name = filters.CharFilter()
    pulp_label_select = LabelSelectFilter()

    class Meta:
        model = Repository
        fields = {"name": NAME_FILTER_OPTIONS}


class BaseRepositoryViewSet(NamedModelViewSet):
    """
    A base class for any repository viewset.
    """

    queryset = Repository.objects.exclude(user_hidden=True).order_by("name")
    serializer_class = RepositorySerializer
    endpoint_name = "repositories"
    router_lookup = "repository"
    filterset_class = RepositoryFilter


class ListRepositoryViewSet(BaseRepositoryViewSet, mixins.ListModelMixin):
    """Endpoint to list all repositories."""

    @classmethod
    def is_master_viewset(cls):
        """Do not hide from the routers."""
        return False


class ReadOnlyRepositoryViewSet(
    BaseRepositoryViewSet,
    mixins.ListModelMixin,
    mixins.RetrieveModelMixin,
):
    """
    A readonly repository which allows only GET method.
    """

    queryset = Repository.objects.all().order_by("name")
    serializer_class = RepositorySerializer
    endpoint_name = "repositories"
    router_lookup = "repository"
    filterset_class = RepositoryFilter


class ImmutableRepositoryViewSet(
    ReadOnlyRepositoryViewSet,
    mixins.CreateModelMixin,
    AsyncRemoveMixin,
):
    """
    An immutable repository ViewSet that does not allow the usage of the methods PATCH and PUT.
    """


class RepositoryViewSet(ImmutableRepositoryViewSet, AsyncUpdateMixin):
    """
    A ViewSet for an ordinary repository.
    """


class RepositoryVersionContentFilter(Filter):
    """
    Filter used to get the repository versions where some given content can be found.
    """

    def __init__(self, *args, **kwargs):
        kwargs.setdefault("help_text", _("Content Unit referenced by HREF"))
        super().__init__(*args, **kwargs)

    def filter(self, qs, value):
        """
        Args:
            qs (django.db.models.query.QuerySet): The RepositoryVersion Queryset
            value (string): of content href to filter

        Returns:
            Queryset of the RepositoryVersions containing the specified content
        """

        if value is None:
            # user didn't supply a value
            return qs

        if not value:
            raise serializers.ValidationError(detail=_("No value supplied for content filter"))

        # Get the content object from the content_href
        content = NamedModelViewSet.get_resource(value, Content)

        return qs.with_content([content.pk])


class RepositoryVersionFilter(BaseFilterSet):
    # e.g.
    # /?number=4
    # /?number__range=4,6
    # /?pulp_created__gte=2018-04-12T19:45
    # /?pulp_created__range=2018-04-12T19:45,2018-04-13T20:00
    # /?content=/pulp/api/v3/content/file/fb8ad2d0-03a8-4e36-a209-77763d4ed16c/
    number = filters.NumberFilter()
    pulp_created = IsoDateTimeFilter()
    content = RepositoryVersionContentFilter()
    content__in = RepositoryVersionContentFilter(field_name="content", lookup_expr="in")

    class Meta:
        model = RepositoryVersion
        fields = {
            "number": ["exact", "lt", "lte", "gt", "gte", "range"],
            "pulp_created": DATETIME_FILTER_OPTIONS,
        }


class RepositoryVersionViewSet(
    NamedModelViewSet, mixins.RetrieveModelMixin, mixins.ListModelMixin, mixins.DestroyModelMixin
):
    endpoint_name = "versions"
    nest_prefix = "repositories"
    router_lookup = "version"
    lookup_field = "number"
    parent_viewset = RepositoryViewSet
    parent_lookup_kwargs = {"repository_pk": "repository__pk"}
    serializer_class = RepositoryVersionSerializer
    queryset = RepositoryVersion.objects.complete()
    filterset_class = RepositoryVersionFilter
    filter_backends = (OrderingFilter, DjangoFilterBackend)
    ordering = ("-number",)

    @extend_schema(
        description="Trigger an asynchronous task to delete a repository version.",
        responses={202: AsyncOperationResponseSerializer},
    )
    def destroy(self, request, repository_pk, number):
        """
        Queues a task to handle deletion of a RepositoryVersion
        """
        version = self.get_object()

        task = dispatch(
            tasks.repository.delete_version,
            exclusive_resources=[version.repository],
            kwargs={"pk": version.pk},
        )
        return OperationPostponedResponse(task, request)

    @extend_schema(
        description="Trigger an asynchronous task to repair a repository version.",
        responses={202: AsyncOperationResponseSerializer},
    )
    @action(detail=True, methods=["post"])
    def repair(self, request, repository_pk, number):
        """
        Queues a task to repair corrupted artifacts corresponding to a RepositoryVersion
        """
        version = self.get_object()
        serializer = RepairSerializer(data=request.data)
        serializer.is_valid()

        verify_checksums = serializer.validated_data["verify_checksums"]

        task = dispatch(
            tasks.repository.repair_version,
            shared_resources=[version.repository],
            args=[version.pk, verify_checksums],
        )
        return OperationPostponedResponse(task, request)


class RemoteFilter(BaseFilterSet):
    """
    Plugin remote filter should:
     - inherit from this class
     - add any specific filters if needed
     - define a `Meta` class which should:
       - specify a plugin remote model for which filter is defined
       - extend `fields` with specific ones
    """

    name = filters.CharFilter()
    pulp_label_select = LabelSelectFilter()
    pulp_last_updated = IsoDateTimeFilter()

    class Meta:
        model = Remote
        fields = {"name": NAME_FILTER_OPTIONS, "pulp_last_updated": DATETIME_FILTER_OPTIONS}


class RemoteViewSet(
    NamedModelViewSet,
    mixins.CreateModelMixin,
    mixins.RetrieveModelMixin,
    mixins.ListModelMixin,
    AsyncUpdateMixin,
    AsyncRemoveMixin,
):
    endpoint_name = "remotes"
    serializer_class = RemoteSerializer
    queryset = Remote.objects.all()
    filterset_class = RemoteFilter


# We have to use GenericViewSet as NamedModelViewSet causes
# get_viewset_for_model() to match multiple viewsets.
class ListRepositoryVersionViewSet(
    GenericViewSet,
    mixins.ListModelMixin,
):
    endpoint_name = "repository_versions"
    serializer_class = RepositoryVersionSerializer
    queryset = RepositoryVersion.objects.exclude(repository__user_hidden=True)
    filterset_class = RepositoryVersionFilter
